import logging
import os.path

from datetime import timedelta
from datetime import datetime

import psutil
from pynvml import *

from .timer import Timer

MACHINE_ACTIVITY_MONITOR = 'machine_activity_monitors'


class UserActivity:
    def __init__(self):
        pass

    def get_last_activity(self):
        """Returns None or a datetime object."""
        return None


class BashHistoryUserActivity(UserActivity):
    def __init__(self):
        UserActivity.__init__(self)

    def get_last_activity(self):
        contents = ''
        try:
            with open(os.path.expanduser('~/.bash_history')) as history_file:
                contents = history_file.read()
        except BaseException as exception:
            return None

        last_timestamp_index = contents.rfind('#')
        if last_timestamp_index < 0:
            return None

        contents = contents[last_timestamp_index + 1:]
        lines = contents.split(os.linesep)
        timestamp = datetime.fromtimestamp(int(lines[0]))

        return timestamp


class GPUMonitor:
    def __init__(self):
        self.alive = True

        try:
            nvmlInit()
        except NVMLError as exception:
            self.alive = False

    def get_gpu_usage(self):
        if not self.alive:
            return []

        usage = []

        device_count = nvmlDeviceGetCount()
        for i in range(device_count):
            handle = nvmlDeviceGetHandleByIndex(i)
            try:
                util = nvmlDeviceGetUtilizationRates(handle)
                usage.append(util.gpu)
            except NVMLError as err:
                continue

        return usage


class MachineActivity:
    def __init__(self, dead_duration=timedelta(hours=1)):
        """
        :param dead_duration: If machine is inactive longer than this time, it is considered "dead".
        :type dead_duration timedelta
        """
        self.dead_duration = dead_duration
        self.inactive_since = None

    def is_active(self):
        """return: true if the machine is considered active."""
        raise NotImplementedError

    def get_duration_inactive(self):
        """return: timedelta of how long machine has been inactive"""
        if self.inactive_since is None:
            return timedelta()

        return self.get_now() - self.inactive_since

    def is_dead(self):
        """:return true if machine has been inactive longer than, or equal to dead_duration"""
        if self.inactive_since is None:
            return False

        duration_inactive = self.get_now() - self.inactive_since
        return duration_inactive >= self.dead_duration

    def check_activity(self):
        """Checks if the machine is currently active. If it is, inactivity timer is reset.
            :return True if active, False otherwise"""
        if self.is_active():
            self.inactive_since = None
            return True

        if self.inactive_since is None:
            self.inactive_since = self.get_now()

        return False

    def get_now(self):
        """Convenience method for getting current time. It is a method so subclasses can override and
            specify timezone, etc, if needed.
        """
        return datetime.now()


class ThresholdMachineActivity(MachineActivity):
    def __init__(self, cpu_threshold=40, gpu_threshold=40, user_activity_threshold=timedelta(minutes=20), dead_duration=timedelta(hours=1)):
        """
        :param cpu_threshold: If all CPU usage is below this percentage, it is considered inactive
        :type cpu_threshold float
        :param gpu_threshold: If GPU (if exits) is below this percentage, it is considered inactive.
        :type gpu_threshold float
        :param user_activity_threshold: If user has not ran a command within this time
        :type user_activity_threshold timedelta
        :param dead_duration: If machine is inactive longer than this time, it is considered "dead".
        :type dead_duration timedelta
        """
        MachineActivity.__init__(self, dead_duration)

        self.cpu_threshold = cpu_threshold
        self.gpu_threshold = gpu_threshold
        self.user_activity_threshold = user_activity_threshold
        self.gpu_monitor = GPUMonitor()
        self.user_activity = BashHistoryUserActivity()
        self.logger = logging.getLogger(MACHINE_ACTIVITY_MONITOR)

    def is_cpu_active(self):
        """:return true if CPU usage is >= cpu_threshold"""

        cpu_usage = psutil.cpu_percent()

        self.logger.info('CPU {} : {}'.format(cpu_usage, self.cpu_threshold))

        return cpu_usage >= self.cpu_threshold

    def is_gpu_active(self):
        """:return true if GPU exists on machine and GPU usage is >= gpu_threshold"""

        i = 0
        for gpu_usage in self.gpu_monitor.get_gpu_usage():
            i += 1
            self.logger.info("GPU {}: {} :{}".format(i, gpu_usage, self.gpu_threshold))
            if gpu_usage >= self.gpu_threshold:
                return True

        return False

    def is_user_active(self):
        """:return true if user has ran a command within the threshold time"""
        latest_timestamp = self.user_activity.get_last_activity()

        if latest_timestamp is None:
            self.logger.info('No timestamp in bash_history')
            return False

        duration = self.get_now() - latest_timestamp

        self.logger.info('Duration from last command {} / {}'.format(duration, self.user_activity_threshold))

        return duration < self.user_activity_threshold

    def is_active(self):
        """return: true if the machine is considered active."""
        active = self.is_cpu_active() or self.is_user_active()

        if not self.gpu_monitor.alive:
            return active

        return active or self.is_gpu_active()

    def get_cpu_stats(self):
        return psutil.cpu_percent()

    def get_gpu_stats(self):
        stats = []
        for gpu_usage in self.gpu_monitor.get_gpu_usage():
            stats.append(gpu_usage)

        return stats

    def get_statistics(self):
        return {
            'cpu': {
                'usage': self.get_cpu_stats(),
                'threshold': self.cpu_threshold,
                'active': self.is_cpu_active()
            },
            'gpu': {
                'usage': self.get_gpu_stats(),
                'threshold': self.gpu_threshold,
                'active': self.is_gpu_active(),
            },
            'user_active': self.is_user_active()
        }


class MachineActivityMonitor(Timer):
    def __init__(self,  activity_monitor, dead_callback):
        """
        :param activity_monitor:
        :type activity_monitor MachineActivity
        """
        # Run every 60 seconds
        Timer.__init__(self, 60)
        self.activity_monitor = activity_monitor
        self.dead_callback = dead_callback

        self.logger = logging.getLogger(MACHINE_ACTIVITY_MONITOR)

    def run(self):
        while not self.stop_flag.wait(self.delay):
            self.logger.info('checking activity, we are active: {}'.format(self.activity_monitor.check_activity()))

            if self.activity_monitor.is_dead() and self.dead_callback is not None:
                self.dead_callback()
                return






"""
Machine types commands
"""

import click

from onepanel.commands.base import APIViewController
from onepanel.commands.login import login_required


class MachineTypesViewController(APIViewController):

    def __init__(self,conn):
        APIViewController.__init__(self,conn)


@click.group(name='machine-types', help='Machine (hardware) commands group')
@click.pass_context
def machine_types(ctx):
    ctx.obj['vc'] = MachineTypesViewController(ctx.obj['connection'])


@machine_types.command('list', help='Show available machine types')
@click.pass_context
@login_required
def list_machine_types(ctx):
    vc = ctx.obj['vc']
    print_data = vc.list(params="/machine_types")
    fields = ['uid','name']
    field_names = ['ID','SPECS']
    empty_message = 'No available machine types.'
    vc.print_items(print_data,fields,field_names,empty_message)




from onepanel.models.api_json import APIJSON
from onepanel.models.account import Account

class Project(APIJSON):
    def __init__(self):
        self.uid = None
        self.account = Account()

    def api_json(self):
        return {
            'uid': self.uid,
        }

    @classmethod
    def from_json(cls, dct):
        project = cls()

        project.uid = dct['uid']
        project.account = Account.from_json(dct['account'])

        return project


import os
import threading
import time
import logging
import configobj

import click
from threading import Thread, Event
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
from onepanel.utilities.timer import Timer
from onepanel.utilities.original_connection import Connection
from onepanel.utilities.s3_authenticator import S3Authenticator
from onepanel.utilities.aws_utility import AWSUtility
from onepanel.utilities.dataset_api import DatasetAPI


LOGGER = 'dataset-downloader'
# This is the directory datasets will be downloaded to if they are aliased.
DATASET_DOWNLOAD_DIRECTORY_IF_ALIAS = '.onepanel-datasets'

class DatasetDownloadStatus:
    """Watches a directory for files downloaded and keeps track of how many were downloaded
        and their statistics, like size.
    """
    def __init__(self, path):
        self.file_count = 0
        self.byte_count = 0
        self.path = path
        self.files_downloading = []

    def add_files(self, files):
        lock = threading.Lock()
        with lock:
            self.file_count += files

    def add_bytes(self, bytes_to_add):
        lock = threading.Lock()
        with lock:
            self.byte_count += bytes_to_add

    def set_files(self, value):
        lock = threading.Lock()
        with lock:
            self.file_count = value

    def set_bytes(self, value):
        lock = threading.Lock()
        with lock:
            self.byte_count = value

    def get_file_count(self):
        return self.file_count

    def get_byte_count(self):
        return self.byte_count

    def get_files_downloading(self):
        return self.files_downloading

    def add_file_downloading(self, file_path):
        lock = threading.Lock()
        with lock:
            self.files_downloading.append(file_path)

    def remove_file_downloading(self, file_path):
        lock = threading.Lock()
        with lock:
            self.files_downloading.remove(file_path)


class DatasetMountIdentifier:
    def __init__(self, account_uid, project_uid, resource_type, resource_uid):
        self.account_uid = account_uid
        self.project_uid = project_uid
        self.resource_type = resource_type
        self.resource_uid = resource_uid
        self.dataset_uid = ''
        self.dataset_version = 0
        self.dataset_mount_uuid = ''
        self.alias = None


class DatasetMountProgressUpdater:
    def __init__(self, api, stats, account_uid, project_uid, dataset_mount_uuid):
        self.stats = stats
        self.api = api
        self.account_uid = account_uid
        self.project_uid = project_uid
        self.dataset_mount_uuid = dataset_mount_uuid

    def update(self):
        total_bytes = self.calculate_total_bytes()
        total_files = self.stats.get_file_count()

        logging.getLogger(LOGGER).info('Sending Update to API files/bytes {}/{}'.format(total_files, total_bytes))
        response = self.api.update_dataset_mount_downloader(
            account_uid=self.account_uid,
            project_uid=self.project_uid,
            dataset_mount_uuid=self.dataset_mount_uuid,
            files_downloaded=total_files,
            bytes_downloaded=total_bytes)

        if response['status_code'] != 200:
            logging.getLogger(LOGGER).warning('Unable to update API. Response code is {}'.format(response['status_code']))

    def calculate_total_bytes(self):
        total_byte_count = self.stats.get_byte_count()

        for file_path in self.stats.get_files_downloading():
            try:
                total_byte_count += os.stat(file_path).st_size
            except:
                pass

        return total_byte_count


class DatasetDownloadStatusPrinter:
    def __init__(self, stats):
        self.stats = stats

    def print_stats(self):
        stats_string = 'Files: {files} Bytes: {bytes}'.format(files=self.stats.get_file_count(), bytes=self.stats.get_byte_count())
        click.echo(stats_string)


class AWSDownloadFileTracker(FileSystemEventHandler):
    def __init__(self, stats):
        self.stats = stats

    def is_temporary_file(self, file_path):
        # With AWS S3 CLI last dot has 8 hexadecimal characters after it.
        if file_path.count('.') == 0:
            return False

        last_dot_index = file_path.rfind('.')
        return len(file_path[last_dot_index + 1:]) == 8

    def on_created(self, event):
        if event.is_directory:
            return

        if self.is_temporary_file(event.src_path):
            self.stats.add_file_downloading(event.src_path)
            return

        # Otherwise, we have a full fledged downloaded file
        self.update_statistic(event.src_path)

    def on_moved(self, event):
        """For AWS, the file is downloaded as a temporary file if it's big.
           When it finishes, it renames it.
        """
        if event.is_directory:
            return

        self.stats.remove_file_downloading(event.src_path)
        self.update_statistic(event.dest_path)

    def update_statistic(self, file_path):
        logging.getLogger(LOGGER).info('Updating statistics with file {}'.format(file_path))

        self.stats.add_files(1)

        try:
            self.stats.add_bytes(os.stat(file_path).st_size)
        except Exception as argument:
            logging.getLogger(LOGGER).warning('Exception trying to update statistics {}'.format(argument))
            # Don't do anything
            pass


class DatasetDownloader(Thread):
    def __init__(self, owner_uid, dataset_mount, path, heartbeat=6.0, timer_delay=5.0):
        Thread.__init__(self)

        self.owner_uid = owner_uid
        self.path = path
        self.stop_flag = Event()
        self.heartbeat = heartbeat
        self.timer_delay = timer_delay
        self.timer = None
        self.dataset_mount = dataset_mount

        self.file_stats = None
        self.observer = None

        self.started = False

        self.api = None
        self.downloader_process = None

    def start(self):
        Thread.start(self)

        conn = Connection()
        conn.load_credentials()

        self.api = DatasetAPI(conn)

        if not os.path.exists(self.path):
            os.makedirs(self.path)

        self.file_stats = DatasetDownloadStatus(self.path)
        self.observer = Observer()
        self.observer.schedule(AWSDownloadFileTracker(self.file_stats), self.path, True)

        (exit_code, process) = self.start_downloading_dataset(conn, self.path, self.dataset_mount.account_uid,
                                                              self.dataset_mount.dataset_uid,
                                                              self.dataset_mount.dataset_version)
        if exit_code != 0:
            logging.getLogger(LOGGER).warning('AWS CLI can not download. Stopping downloader.')
            self.stop_flag.set()
            self.observer = None
            return None

        self.downloader_process = process
        self.start_monitoring_files()
        self.start_updating_api(self.owner_uid, self.dataset_mount.project_uid,
                                self.dataset_mount.dataset_mount_uuid)

        self.started = True

    def run(self):
        while not self.stop_flag.wait(self.heartbeat):
            if self.started:
                if self.downloader_process.poll() is not None:
                    self.finished_downloading_hook()
                    self.stop()

    def finished_downloading_hook(self):
        response = self.api.finish_dataset_mount_downloader(self.owner_uid, self.dataset_mount.project_uid,
                                                            self.dataset_mount.dataset_mount_uuid)

        response_code = response['status_code']
        if response_code != 200:
            logging.getLogger(LOGGER).warning('Error from finish mount: {}'.format(response_code))

        onepanel_dir = os.path.join(self.path, '.onepanel')
        if not os.path.exists(onepanel_dir):
            os.makedirs(onepanel_dir)

        dataset_file = os.path.join(self.path, os.path.join('.onepanel','dataset'))

        cfg = configobj.ConfigObj(dataset_file)
        cfg['uid'] = self.dataset_mount.dataset_uid
        cfg['account_uid'] = self.owner_uid
        cfg.write()

        if self.dataset_mount.alias is not None:
            symlink_path = self.dataset_mount.alias
            os.symlink(self.path, symlink_path, True)
            logging.getLogger(LOGGER).info('Created symlink {} -> {}'.format(self.path, symlink_path))

    def join(self, timeout=None):
        if self.downloader_process is not None:
            self.downloader_process.wait()

        if self.timer is not None:
            self.timer.join()

        if self.observer is not None:
            self.observer.join()

        Thread.join(self, timeout)

    def stop(self):
        if self.timer is not None:
            self.timer.stop()

        if self.observer is not None:
            self.observer.stop()

        self.stop_flag.set()

    def start_monitoring_files(self):
        self.observer.start()

    def start_updating_api(self, account_uid, project_uid, dataset_mount_uuid):
        updater = DatasetMountProgressUpdater(self.api, self.file_stats, account_uid, project_uid, dataset_mount_uuid)
        self.timer = Timer(self.timer_delay, updater.update)
        self.timer.start()

    def start_downloading_dataset(self, connection, path, account_uid, dataset_uid, dataset_version):
        # Authenticate for S3
        s3auth = S3Authenticator(connection)
        creds = s3auth.get_s3_credentials(account_uid, 'datasets', dataset_uid)
        if creds['data'] is None:
            logging.getLogger(LOGGER).warning("Unable to get S3 credentials. Exiting.")
            return -1
        aws_access_key_id = creds['data']['AccessKeyID']
        aws_secret_access_key = creds['data']['SecretAccessKey']
        aws_session_token = creds['data']['SessionToken']
        aws_util = AWSUtility(aws_access_key_id=aws_access_key_id,
                              aws_secret_access_key=aws_secret_access_key,
                              aws_session_token=aws_session_token)

        aws_util.suppress_output = True

        current_version = self.api.get_dataset_version(account_uid, dataset_uid, dataset_version)

        if current_version['data'] is None:
            if current_version['status_code'] == 404:
                logging.getLogger(LOGGER).info('Dataset is without files.')
                return -1
            else:
                logging.getLogger(LOGGER).warning('Could not get information about dataset.')
                return -1

        if current_version['data']['provider']['uid'] != 'aws-s3':
            logging.getLogger(LOGGER).warning('Unsupported dataset provider')
            return -1

        s3_from_dir = current_version['data']['version']['path']
        aws_util.run_cmd_background = True

        logging.getLogger(LOGGER).info('starting AWS download')

        return aws_util.download_all_background(path, s3_from_dir)


class DatasetDownloadListener(Thread):
    """
        continue_listening: Whether the thread should continue waiting and check for new dataset mounts.
    """
    def __init__(self, connection, download_path, account_uid, project_uid,
                 resource_type, resource_uid, continue_listening=True, verbose=False):
        Thread.__init__(self)

        self.identifier = DatasetMountIdentifier(account_uid, project_uid, resource_type, resource_uid)
        self.api = DatasetAPI(connection)
        self.download_path = download_path
        self.continue_listening = continue_listening

        if verbose:
            logger = logging.getLogger(LOGGER)
            logger.setLevel(logging.DEBUG)
            ch = logging.StreamHandler()
            ch.setLevel(logging.DEBUG)
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            ch.setFormatter(formatter)
            logger.addHandler(ch)

    def get_dataset_mount(self):
        if self.identifier.resource_type == 'instance':
            return self.api.claim_instance_mount(
                        account_uid=self.identifier.account_uid,
                        project_uid=self.identifier.project_uid,
                        instance_uid=self.identifier.resource_uid,
                        downloader_id=os.getcwd(),
                        downloader_pid=os.getpid())
        elif self.identifier.resource_type == 'job':
            return self.api.claim_job_mount(
                        account_uid=self.identifier.account_uid,
                        project_uid=self.identifier.project_uid,
                        job_uid=self.identifier.resource_uid,
                        downloader_id=os.getcwd(),
                        downloader_pid=os.getpid())
        else:
            raise ValueError('Resource Type {} not supported'.format(self.identifier.resource_type))

    def run(self):
        logger = logging.getLogger(LOGGER)

        while True:
            logger.info('Listening for Dataset Mounts')

            try:
                response = self.get_dataset_mount()
            except Exception as Argument:
                # On exception, try again in 10 seconds
                logger.warning('Got an exception trying to claim an instance mount {}. Trying again in 10 seconds'.format(Argument))

                time.sleep(10)
                continue

            response_code = response['status_code']
            if response_code != 200:
                if response_code != 404:
                    logger.warning('Error from dataset mount: {}'.format(response_code))
                elif response_code == 404:
                    if not self.continue_listening:
                        logger.info('No more datasets to mount. Done.')
                        return
                    logger.info('No datasets to mount. Checking again in 30 seconds')
                # API says there is no more. Try back later.
                time.sleep(30)
                continue

            dataset_mount = response['data']

            download_identifier = DatasetMountIdentifier(dataset_mount['dataset']['account']['uid'],
                                                         self.identifier.project_uid,
                                                         self.identifier.resource_type,
                                                         self.identifier.resource_uid)

            download_identifier.dataset_mount_uuid = dataset_mount['uuid']
            download_identifier.dataset_uid = dataset_mount['dataset']['uid']
            download_identifier.dataset_version = dataset_mount['dataset_versioning']['version']

            path = os.path.join(self.download_path,
                                download_identifier.account_uid,
                                download_identifier.dataset_uid,
                                str(download_identifier.dataset_version))

            if dataset_mount['alias']:
                download_identifier.alias = os.path.join(self.download_path, dataset_mount['alias'])
                path = os.path.join(self.download_path, 
                                DATASET_DOWNLOAD_DIRECTORY_IF_ALIAS,
                                download_identifier.account_uid,
                                download_identifier.dataset_uid,
                                str(download_identifier.dataset_version))

            logger.info('Obtained dataset mount. Downloading to {}'.format(path))

            downloader = DatasetDownloader(self.identifier.account_uid, download_identifier, path)

            downloader.start()
            downloader.join()

            logger.info('Dataset {} Finished downloading'.format(path))




from onepanel.models.api_json import APIJSON


class MachineType(APIJSON):
    def __init__(self):
        self.uid = None
        self.info = {}

    def api_json(self):
        return {
            'uid': self.uid
        }

    @classmethod
    def from_json(cls, dct):
        machine = cls()

        machine.uid = dct['uid']
        machine.info['cpu'] = dct['info']['cpu']
        machine.info['gpu'] = dct['info'].get('gpu')
        machine.info['ram'] = dct['info']['ram']

        return machine



import platform
import subprocess
import os
import sys
from distutils.sysconfig import get_python_lib

from onepanel.utilities.s3.aws_cli_wrapper import AWSCLIWrapper


class AWSUtility:
	env = {} # Use for shell command environment variables
	suppress_output = False
	run_cmd_background = False
	# Windows Specific
	# https://docs.microsoft.com/en-us/windows/desktop/ProcThread/process-creation-flags
	# This is used to run processes in the background on Windows
	CREATE_NO_WINDOW = 0x08000000

	def __init__(self,aws_access_key_id='',
				 aws_secret_access_key='',
				 aws_session_token=''):
		self.aws_access_key_id = aws_access_key_id
		self.aws_secret_access_key = aws_secret_access_key
		self.aws_session_token = aws_session_token
		self.env = {'AWS_ACCESS_KEY_ID': self.aws_access_key_id, 'AWS_SECRET_ACCESS_KEY': self.aws_secret_access_key,
			   'AWS_SESSION_TOKEN': self.aws_session_token,
		}
		if platform.system() is 'Windows':
			self.env[str('SYSTEMROOT')] = os.environ['SYSTEMROOT']
			self.env[str('PATH')] = os.environ['PATH']
			self.env[str('PYTHONPATH')] = os.pathsep.join(sys.path)

	def build_full_s3_url(self,s3_path):
		s3_path = 's3://{bucket}/{path}'.format(bucket=self.get_dataset_bucket_name(),path=s3_path)
		return s3_path

	def get_dataset_bucket_name(self):
		return os.getenv('DATASET_BUCKET', 'onepanel-datasets')

	def upload_dir(self, dataset_directory, s3_directory, exclude='', delete_removed=False, authenticator=None):
		s3_path = 's3://{bucket}/{path}'.format(bucket=self.get_dataset_bucket_name(), path=s3_directory)
		aws_full_path = self.get_full_path_to_aws_cli()
		if "aws" not in aws_full_path:
			return -1

		cmd_list = []
		close_fds = False
		exclude_arg = '--exclude "{ex_str}"'.format(ex_str=exclude)
		cmd_list = cmd_list + [aws_full_path, 's3', 'sync', dataset_directory, s3_path,exclude_arg]

		if delete_removed:
			cmd_list.append('--delete')

		# Need to pass the command as one long string. Passing in a list does not work when executed.
		cmd = ' '.join(cmd_list)
		print("Uploading...")

		wrapper = AWSCLIWrapper(authenticator)
		results = wrapper.run(args=cmd, env=self.env, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
							  stderr=subprocess.PIPE, shell=True, close_fds=close_fds)

		if self.suppress_output is False:
			suppress_s3_info = s3_path.rsplit('/', 1)[0]

			cleaned_output = results[1].replace(suppress_s3_info, '')
			sys.stdout.write(cleaned_output)

			cleaned_err = results[2].replace(suppress_s3_info, '')
			sys.stdout.write(cleaned_err)

		return 0


	def download_all(self, dataset_directory, s3_directory, delete=False):
		s3_path = 's3://{bucket}/{path}'.format(bucket=self.get_dataset_bucket_name(), path=s3_directory)
		aws_full_path = self.get_full_path_to_aws_cli()
		if "aws" in aws_full_path:
			cmd_list = []
			close_fds = False
			if self.run_cmd_background:
				if sys.platform != 'win32':
					cmd_list.insert(0,'nice')
					cmd_list.insert(0,'nohup')
					close_fds = True
				else:
					# /i so that windows doesn't create "%SYSTEM_DRIVE%" folder
					cmd_list.insert(0, 'start /b /i')

			cmd_list = cmd_list + [aws_full_path, 's3', 'sync', s3_path, dataset_directory]
			if delete:
				cmd_list.append('--delete')
				cmd_list.append('--exclude ".onepanel/*"')
			# Need to pass the command as one long string. Passing in a list does not work when executed.
			cmd = ' '.join(cmd_list)
			if self.run_cmd_background:
				print("Download starting in background.")
			else:
				print("Downloading...")

			# shell=True because we want to intercept the output from the command.
			# And also, it fixes issues with executing the string of commands.
			if self.run_cmd_background:
				if sys.platform != 'win32':
					subprocess.Popen(args=cmd, env=self.env, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
								 stderr=subprocess.PIPE, shell=True, close_fds=close_fds)
				else:
					subprocess.Popen(args=cmd, env=self.env, shell=True, close_fds=close_fds,
									 creationflags=self.CREATE_NO_WINDOW)
			else:
				p = subprocess.Popen(args=cmd, env=self.env, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
							 stderr=subprocess.PIPE, shell=True,close_fds=close_fds)
				suppress_s3_info = s3_path.rsplit('/', 1)[0]
				for line in iter(p.stdout.readline, '' or b''):  # replace '' with b'' for Python 3
					if self.suppress_output is False:
						line_str = line.decode()
						cleaned_line = line_str.replace(suppress_s3_info, '')
						sys.stdout.write(cleaned_line)
			return 0
		return -1

	# TODO temporary method - need to redefine interface for above method
	def download_all_background(self, dataset_directory, s3_directory):
		s3_path = 's3://{bucket}/{path}'.format(bucket=self.get_dataset_bucket_name(), path=s3_directory)
		aws_full_path = self.get_full_path_to_aws_cli()

		if 'aws' not in aws_full_path:
			return -1

		cmd_list = []
		close_fds = False
		if sys.platform != 'win32':
			cmd_list.insert(0, 'nice')
			cmd_list.insert(0, 'nohup')
			close_fds = True
		else:
			# /i so that windows doesn't create "%SYSTEM_DRIVE%" folder
			cmd_list.insert(0, 'start /b /i')
		cmd_list = cmd_list + [aws_full_path, 's3', 'sync', s3_path, dataset_directory]
		# Need to pass the command as one long string. Passing in a list does not work when executed.
		cmd = ' '.join(cmd_list)

		# shell=True because we want to intercept the output from the command.
		# And also, it fixes issues with executing the string of commands.
		if sys.platform != 'win32':
			p = subprocess.Popen(args=cmd, env=self.env, stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL,
						 		 stderr=subprocess.DEVNULL, shell=True, close_fds=close_fds)

			return 0, p
		else:
			p = subprocess.Popen(args=cmd, env=self.env, shell=True, close_fds=close_fds,
							 	creationflags=self.CREATE_NO_WINDOW)

			return 0, p

	def download(self,to_dir,s3_full_path_to_file):
		s3_path = 's3://{bucket}/{path}'.format(bucket=self.get_dataset_bucket_name(), path=s3_full_path_to_file)
		aws_full_path = self.get_full_path_to_aws_cli()
		if "aws" in aws_full_path:
			cmd_list = []
			close_fds = False
			if self.run_cmd_background:
				if sys.platform != 'win32':
					cmd_list.insert(0,'nice')
					cmd_list.insert(0,'nohup')
					close_fds = True
				else:
					# /i so that windows doesn't create "%SYSTEM_DRIVE%" folder
					cmd_list.insert(0, 'start /b /i')
			cmd_list = cmd_list + [aws_full_path,'s3', 'cp', s3_path, to_dir]
			# Need to pass the command as one long string. Passing in a list does not work when executed.
			cmd = ' '.join(cmd_list)
			if self.run_cmd_background:
				print("Download starting in background.")
				if sys.platform != 'win32':
					subprocess.Popen(args=cmd, env=self.env, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
									 stderr=subprocess.PIPE, shell=True, close_fds=close_fds)
				else:
					subprocess.Popen(args=cmd, env=self.env, shell=True, close_fds=close_fds,
									 creationflags=self.CREATE_NO_WINDOW)
			else:
				print("Downloading...")
				# Also need to execute through the shell
				p = subprocess.Popen(args=cmd, env=self.env, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
								 stderr=subprocess.PIPE, shell=True, close_fds=close_fds)
				suppress_s3_info = s3_path.rsplit('/',1)[0]
				for line in iter(p.stdout.readline, '' or b''):  # replace '' with b'' for Python 3
					if self.suppress_output is False:
						line_str = line.decode()
						cleaned_line = line_str.replace(suppress_s3_info,'')
						sys.stdout.write(cleaned_line)
			return 0
		return -1

	def check_s3_path_for_files(self,full_s3_path='',recursive=True):
		ret_val = {'data':None, 'code':-1, 'msg':''}
		if full_s3_path == '':
			ret_val = {'data': None, 'code': -1, 'msg': 'Need the full s3 path passed in.'}
			return ret_val
		aws_full_path = self.get_full_path_to_aws_cli()
		if "aws" in aws_full_path:
			recursive_arg = ''
			if recursive:
				recursive_arg = '--recursive'
			# Need to pass the command as one long string. Passing in a list does not work when executed.
			cmd = ' '.join([aws_full_path, 's3', 'ls',recursive_arg,'--summarize',full_s3_path])
			# Also need to execute through the shell
			p = subprocess.Popen(args=cmd, env=self.env,stdin=subprocess.PIPE, stdout=subprocess.PIPE,
							 stderr=subprocess.PIPE, shell=True)
			p.wait()
			raw_output = p.stdout.readlines()
			for line in raw_output:
				if 'Total Objects' in line.decode():
					objects_str = line.decode()
					str_split = objects_str.split(':')
					# Grab the string of the number
					string_num = str_split[-1].strip()
					ret_val = {'data': int(string_num), 'code': 0, 'msg': 'Total files found.'}
					break
			return ret_val
		return ret_val

	def get_s3_path_details(self,full_s3_path='',total_files = True,total_bytes = True):
		ret_val = {'data': None, 'code': -1, 'msg': ''}
		data = {}
		if full_s3_path == '':
			ret_val = {'data': None, 'code': -1, 'msg': 'Need the full s3 path passed in.'}
			return ret_val
		aws_full_path = self.get_full_path_to_aws_cli()
		if "aws" in aws_full_path:
			# Need to pass the command as one long string. Passing in a list does not work when executed.
			cmd = ' '.join([aws_full_path, 's3api', 'list-objects','--bucket',self.get_dataset_bucket_name(),
							'--prefix',full_s3_path,'--output','json','--query',
							'"[sum(Contents[].Size), length(Contents[])]"'])
			# Also need to execute through the shell
			p = subprocess.Popen(args=cmd, env=self.env, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
								 stderr=subprocess.PIPE, shell=True)
			p.wait()
			raw_output = p.stdout.readlines()
			# This can happen if the updater thread calls this before the upload starts
			# There will not be any output at that point, we don't want it to try
			if len(raw_output) < 1:
				data['total_bytes'] = "0"
				data['total_files'] = "0"
				ret_val = {'data': data, 'code': 0, 'msg': 'Data found.'}
				return ret_val
			if total_bytes:
				total_bytes_str = raw_output[1].decode("utf-8").strip().strip(',')
				data['total_bytes'] = total_bytes_str
			if total_files:
				total_files_str = raw_output[2].decode("utf-8").strip().strip(',')
				data['total_files'] = total_files_str

			ret_val = {'data': data, 'code': 0, 'msg': 'Data found.'}
		return ret_val

	def get_full_path_to_aws_cli(self):
		# Figure out the full path to awscli
		if platform.system() is 'Windows':
			path_to_aws_cmd = ['where','aws']
		else:
			path_to_aws_cmd = ['which', 'aws']
		p = subprocess.Popen(args=path_to_aws_cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
							 stderr=subprocess.PIPE, shell=False)
		p.wait()
		line = p.stdout.readline()
		return line.decode().rstrip()



import json
from onepanel.commands.base import APIViewController

class DatasetAPI(APIViewController):
    def __init__(self, connection):
        APIViewController.__init__(self, connection)
        self.connection = self.conn

    def claim_instance_mount(self, account_uid, project_uid, instance_uid, downloader_id, downloader_pid):
        endpoint = '/accounts/{account_uid}/projects/{project_uid}/instances/{instance_uid}/claim_dataset_mount'.format(
            account_uid=account_uid,
            project_uid=project_uid,
            instance_uid=instance_uid
        )

        post = {
            'downloaderID': downloader_id,
            'downloaderPID': downloader_pid
        }

        return self.put(params=endpoint, post_object=post)

    def claim_job_mount(self, account_uid, project_uid, job_uid, downloader_id, downloader_pid):
        endpoint = '/accounts/{account_uid}/projects/{project_uid}/jobs/{job_uid}/claim_dataset_mount'.format(
            account_uid=account_uid,
            project_uid=project_uid,
            job_uid=job_uid
        )

        post = {
            'downloaderID': downloader_id,
            'downloaderPID': downloader_pid
        }

        return self.put(params=endpoint, post_object=post)

    def get_dataset_version(self, account_uid, dataset_uid, version=None):
        if version is None:
            version = 'current'

        endpoint = '/accounts/{account_uid}/datasets/{dataset_uid}/version/{version}'.format(
            account_uid=account_uid,
            dataset_uid=dataset_uid,
            version=version
        )
        return self.get(params=endpoint)

    def update_dataset_mount_downloader(self, account_uid, project_uid, dataset_mount_uuid, files_downloaded, bytes_downloaded):
        endpoint = '/accounts/{account_uid}/projects/{project_uid}/dataset_mounts/{dataset_mount_uuid}/update'.format(
            account_uid=account_uid,
            project_uid=project_uid,
            dataset_mount_uuid=dataset_mount_uuid
        )

        post = {
            'filesDownloaded': files_downloaded,
            'bytesDownloaded': bytes_downloaded
        }

        return self.put(params=endpoint, post_object=post)

    def finish_dataset_mount_downloader(self, account_uid, project_uid, dataset_mount_uuid):
        endpoint = '/accounts/{account_uid}/projects/{project_uid}/dataset_mounts/{dataset_mount_uuid}/finish'.format(
            account_uid=account_uid,
            project_uid=project_uid,
            dataset_mount_uuid=dataset_mount_uuid
        )

        return self.put(params=endpoint)

    def get_dataset_mount(self, account_uid, project_uid, dataset_mount_uuid):
        endpoint = '/accounts/{account_uid}/projects/{project_uid}/dataset_mounts/{dataset_mount_uuid}'.format(
            account_uid=account_uid,
            project_uid=project_uid,
            dataset_mount_uuid=dataset_mount_uuid
        )

        return self.get(params=endpoint)

    def delete_dataset_mount(self, account_uid, project_uid, dataset_mount_uuid):
        endpoint = '/accounts/{account_uid}/projects/{project_uid}/dataset_mounts/{dataset_mount_uuid}'.format(
            account_uid=account_uid,
            project_uid=project_uid,
            dataset_mount_uuid=dataset_mount_uuid
        )

        return self.delete(uid=endpoint)


import json

from onepanel.models.api_json import APIJSON

class APIJSONEncoder(json.JSONEncoder):
    """JSONEncoder for uploading items via API"""
    def default(self, obj):
        if isinstance(obj, APIJSON):
            return obj.api_json()

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


import os

import functools
import requests
from configparser import ConfigParser


class Connection:
    """ REST API requests defaults and credentials
    """
    def __init__(self):
        self.URL = os.getenv('BASE_API_URL', 'https://c.onepanel.io/api')
        self.SSL_VERIFY = True
        self.headers = {'Content-Type': 'application/json'}
        self.account_uid = None
        self.user_uid = None
        self.token = None
        self.gitlab_impersonation_token = None

        # wrap requests methods to reduce number of arguments in api queries
        self.get = functools.partial(requests.get, headers=self.headers, verify=self.SSL_VERIFY)
        self.post = functools.partial(requests.post, headers=self.headers, verify=self.SSL_VERIFY)
        self.put = functools.partial(requests.put, headers=self.headers, verify=self.SSL_VERIFY)
        self.delete = functools.partial(requests.delete, headers=self.headers, verify=self.SSL_VERIFY)
        self.head = functools.partial(requests.head, headers=self.headers, verify=self.SSL_VERIFY)

    def save_credentials(self, data):
        credentials = ConfigParser()
        credentials['default'] = {'uid': data['uid'],
                                  'token': data['sessions'][0]['token'],
                                  'gitlab_impersonation_token': data['gitlab_impersonation_token'],
                                  'account_uid': data['account']['uid']}

        onepanel_home = os.path.expanduser(os.path.join('~', '.onepanel'))
        if not os.path.exists(onepanel_home):
            os.makedirs(onepanel_home)

        filename = os.path.join(onepanel_home, 'credentials')
        with open(filename, 'w') as f:
            credentials.write(f)

    def load_credentials(self):
        credentials = ConfigParser()
        filename = os.path.expanduser(os.path.join('~', '.onepanel', 'credentials'))
        credentials.read(filename)

        self.user_uid = credentials.get('default','uid',fallback=None)
        self.account_uid = credentials.get('default', 'account_uid', fallback=None)
        self.token = credentials.get('default', 'token', fallback=None)
        self.gitlab_impersonation_token = credentials.get('default','gitlab_impersonation_token', fallback=None)

        if self.token:
            self.headers['Authorization'] = 'Bearer {}'.format(self.token)

    def clear_credentials(self):
        filename = os.path.expanduser(os.path.join('~', '.onepanel', 'credentials'))

        if not os.path.exists(filename):
            return

        os.remove(filename)



""" Command line interface for the OnePanel Machine Learning platform

Wrap git commands and provide transparent integration onepanel commands with the git.
"""

import glob
import os
import subprocess
import onepanel


class GitWrapper:
    def __init__(self):
        self.HOST = os.getenv('GITLAB_GIT_HOST', 'git.onepanel.io')

    def init(self,onepanel_user_uid,gitlab_token,home, account_uid, project_uid):
        git_cmd = 'git init'
        p = subprocess.Popen(git_cmd, cwd=home, shell=True)
        p.wait()

        git_cmd = 'git remote add onepanel https://{}:{}@{}/{}/{}.git'.format(
            onepanel_user_uid,
            gitlab_token,
            self.HOST,
            account_uid,
            project_uid
        )
        p = subprocess.Popen(git_cmd, cwd=home, shell=True)
        p.wait()
        self.git_hook_code(home)

    def clone(self,onepanel_user_uid,gitlab_token, home, account_uid, project_uid, ext=''):
        git_cmd = 'git clone -o onepanel https://{}:{}@{}/{}/{}{}.git {}'.format(
            onepanel_user_uid,
            gitlab_token,
            self.HOST,
            account_uid,
            project_uid,
            ext,
            home
        )
        p = subprocess.Popen(git_cmd, shell=True)
        p.wait()

        return p.returncode

    def exclude(self, home, files):
        exclude_file = os.path.join(home, '.git/info/exclude')
        with open(exclude_file,'a+') as f:
            for file in files:
                f.write(file + '\n')

    def push(self, home):
        self.git_hook_code(home)
        if len(glob.glob(os.path.join(home, '*'))) < 1:
            print("Cannot proceed with operation, no files to push up.")
            exit(0)

        # The --progress flag is needed, otherwise the progress of the git push
        # isn't reported in the sterr.
        git_cmd = 'git push --progress -u onepanel master'

        p = subprocess.Popen(git_cmd, cwd=home, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        p.wait()

        err = p.stderr.read().decode()
        output = p.stdout.read().decode()
        if 'error: src refspec master does not match any' in err:
            print('Cannot proceed with operation.\nYou first have to add and commit files' 
                  'using git before running onepanel push')
            return

        # Error contains progress, so print that first.
        print(err)

        # Then print output
        print(output)

    def pull(self, home):
        git_cmd = 'git pull onepanel master'
        p = subprocess.Popen(git_cmd, cwd=home, shell=True)
        p.wait()
        self.git_hook_code(home)

    def git_hook_code(self,cwd):
        git_hook_utility = onepanel.utilities.git_hook_utility.GitHookUtility()
        if git_hook_utility.safe_to_add_git_hooks(cwd):
            if git_hook_utility.check_pre_push_gitlab_update_hook(cwd) is False:
                git_hook_utility.add_pre_push_gitlab_update_hook(cwd)
            if git_hook_utility.check_pre_commit_hook(cwd) is False:
                git_hook_utility.add_pre_commit_hook(cwd)


import time
from threading import Thread, Event


class Timer(Thread):
    """A Timer repeats an action until stopped. The action is passed in and
        starts on a separate thread on every invocation. This is to prevent the duration
        of the action from interfering with the timing.
    """
    def __init__(self, delay, action=None):
        Thread.__init__(self)
        self.delay = delay
        self.stop_flag = Event()
        self.callback = action
        
    def run(self):
        while not self.stop_flag.wait(self.delay):
            if self.callback is not None:
                self.fire()
        
    def stop(self):
        self.stop_flag.set()

    def fire(self):
        thread = Thread(target=self.callback)
        thread.start()


if __name__ == '__main__':
    # Sample program
    def sayHi():
        print('hi')

    # Create a timer that executes something every second
    timer = Timer(1, sayHi)
    timer.start()
    
    # Main thread sleep for 3 seconds so the timer can execute a few times
    time.sleep(3)
    
    # Stop the timer
    timer.stop()
    
    # Wait for the timer to end along with this thread.
    timer.join()


""" Command line interface for the OnePanel Machine Learning platform

Entry point for command line interface.
"""


import os

import click

from onepanel.commands.common import clone, download, push, pull, background_dataset_push, timer_sync_output
from onepanel.utilities.original_connection import Connection

from onepanel.commands.datasets import datasets
from onepanel.commands.environments import environments
from onepanel.commands.instances import workspaces
from onepanel.commands.jobs import jobs
from onepanel.commands.login import login, login_with_token, logout
from onepanel.commands.machine_types import machine_types
from onepanel.commands.projects import projects
from onepanel.commands.volume_types import volume_types
from onepanel.constants import *
from onepanel.utilities.git_hook_utility import GitHookUtility
from onepanel.utilities.git_utility import CheckGitLfsInstalled


@click.group()
@click.version_option(version=CLI_VERSION, prog_name='Onepanel CLI')
@click.pass_context
def cli(ctx):
    conn = Connection()
    conn.load_credentials()
    utility = CheckGitLfsInstalled()

    ctx.obj['connection'] = conn
    ctx.obj['git_utility'] = utility
    utility.figure_out_git_installed()
    if utility.git_installed is False:
        print('Error. Cannot detect git, please verify git is installed.')
        exit(-1)


# Ensure that our git-hook to ping an end-point exits
# Are we in a OnePanel dataset (or project), and is the .git folder available in the current dir?
cwd = os.getcwd()
git_hook_utility = GitHookUtility()
if git_hook_utility.safe_to_add_git_hooks(cwd):
    if git_hook_utility.check_pre_push_gitlab_update_hook(cwd) is False:
        git_hook_utility.add_pre_push_gitlab_update_hook(cwd)
    if git_hook_utility.check_pre_commit_hook(cwd) is False:
        git_hook_utility.add_pre_commit_hook(cwd)

cli.add_command(login)
cli.add_command(login_with_token)
cli.add_command(logout)
cli.add_command(clone)
cli.add_command(download)
cli.add_command(push)
cli.add_command(pull)
cli.add_command(projects)
cli.add_command(datasets)
cli.add_command(background_dataset_push)
cli.add_command(jobs)
cli.add_command(machine_types)
cli.add_command(environments)
cli.add_command(volume_types)
cli.add_command(workspaces)
cli.add_command(timer_sync_output)


def main():
    return cli(obj={})


if __name__ == '__main__':
    cli(obj={})



class Credentials:
    def __init__(self, access_key_id, secret_access_key, session_token):
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key
        self.session_token = session_token

    def dict(self):
        return {
            'AWS_ACCESS_KEY_ID': self.access_key_id,
            'AWS_SECRET_ACCESS_KEY': self.secret_access_key,
            'AWS_SESSION_TOKEN': self.session_token
        }


class Provider:
    def __init__(self):
        # Do nothing
        pass

    def loads_credentials(self):
        """Returns true if the provider needs to load the credentials
           and provides them. E.g. if you need to make a network request
           for them. Returns false if you want to use the default logic
           to check for credentials via AWS config files, etc."""

        return False

    def credentials(self):
        return None


class APIProvider(Provider):
    """Loads credentials with a network request"""
    def __init__(self, conn, endpoint):
        Provider.__init__(self)
        self.conn = conn
        self.endpoint = endpoint

    def loads_credentials(self):
        return True

    def credentials(self):
        response = self.conn.get(self.endpoint)

        if response.status_code != 200:
            raise ValueError(response.status_code)

        data = response.json()

        aws_access_key_id = data['AccessKeyID']
        aws_secret_access_key = data['SecretAccessKey']
        aws_session_token = data['SessionToken']

        return Credentials(aws_access_key_id, aws_secret_access_key, aws_session_token)


class StaticProvider(Provider):
    """Provides credentials that are set on object creation"""
    def __init__(self, access_key, secret_access_key, session_token):
        Provider.__init__(self)
        self.access_key = access_key
        self.secret_access_key = secret_access_key
        self.session_token = session_token

    def loads_credentials(self):
        return True

    def credentials(self):
        return Credentials(self.access_key, self.secret_access_key, self.session_token)



import os
import subprocess
import sys

from onepanel.utilities.connection import Connection
from onepanel.commands.projects import ProjectViewController

# The maximum file-size for a file to be committed:
max_file_size = 50 * 1024  # in KiB (= 1024 byte), 50 MB
max_repo_size = 500 * 1024  # in KiB, 200 MB
units = 1024.0  # To move up and down between magnitudes of size

class PrePushToEndpointHook:

    version = None

    def __init__(self):
        self.version = "1.0.0"

    @staticmethod
    def check_for_large_files():
        # Now, do the checking:
        try:
            # Check if the repository size is too big
            git_repo_info = subprocess.check_output(
                ['git', "count-objects", "-v"],
                stderr=subprocess.STDOUT).decode("utf-8")
            repo_info_output = git_repo_info.splitlines()
            for git_output in repo_info_output:
                if "size" in git_output:
                    size_line = git_output
                    kib_info = size_line.split(":")
                    kib_val = int(kib_info[1].strip())
                    if (kib_val * units) >= max_repo_size * units:
                        print(
                            "The total size of your repository is greater than {repo_max_size}.".format(
                                repo_max_size=sizeof_fmt(max_repo_size * units)))
                        print("Use Onepanel datasets to store datasets or any files larger than {file_max_size}.".
                              format(file_max_size=sizeof_fmt(max_file_size * units)))
                        print("Remove any large files and bring your total repo size to under {repo_max_size} and try again.".
                            format(repo_max_size=sizeof_fmt(max_repo_size * units)))
                        sys.exit(1)
            sys.exit(0)

        except subprocess.CalledProcessError:
            # There was a problem calling "git status".
            print("Error executing git status")
            sys.exit(12)

    @staticmethod
    def ping_endpoint():
        conn = Connection()
        conn.load_credentials()
        data = {
            'version':'1',
            'data':''
        }

        # Figure out if this is a project or dataset push
        home = os.getcwd()
        onepanel_dir = os.path.join(home, '.onepanel')
        if not os.path.exists(onepanel_dir):
            print("ERROR.Directory does not exist, cannot connect to endpoint.")
            print("DETAILS." + onepanel_dir)
            exit(-1)

        project_push = True
        project_file = os.path.join(home, ProjectViewController.PROJECT_FILE)
        if not os.path.isfile(project_file):
            project_push = False
        else:
            # Safe to ping the endpoint with project credentials
            pvc = ProjectViewController(conn)
            pvc.init_credentials_retrieval()
            url_params = '/gitlab/accounts/{account_uid}/project/{project_uid}/update'.format(
                account_uid=pvc.account_uid,project_uid=pvc.project_uid
            )
            try:
                pvc.post(data,params=url_params)
            except ValueError as e:
                print("Gitlab Update Failed.")

        if project_push == False:
            print("ERROR. Cannot connect to endpoint, unclear if this code is initialized with 'onepanel init'.")
            exit(-1)

def sizeof_fmt(num):
    """
	This function will return a human-readable filesize-string
	 like "3.5 MB" for it's given 'num'-parameter.
	From http://stackoverflow.com/questions/1094841
	"""
    for x in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if num < units:
            return "%3.1f %s" % (num, x)
        num /= units


import subprocess

from onepanel.utilities.s3.authentication import Provider


class AWSCLIWrapper:
    """Wraps AWS CLI commands so we get new credentials if they expired"""
    def __init__(self, credentials_provider=None, retry=3):
        """
        :param credentials_provider: Provides credentials for requests.
        :type credentials_provider: onepanel.lib.s3.authentication.Provider
        """
        if credentials_provider is None:
            credentials_provider = Provider()

        self.credentials_provider = credentials_provider
        self.retry = retry
        self.retries = 0

    def run(self, *args, **kwargs):
        while self.retries < self.retry:
            p = subprocess.Popen(*args, **kwargs)

            output = ""
            err = ""

            for line in iter(p.stdout.readline, '' or b''):  # replace '' with b'' for Python 3
                output += line.decode()
            for line in iter(p.stderr.readline, '' or b''):  # replace '' with b'' for Python 3
                err += line.decode()

            if ('(ExpiredToken)' in err or '(InvalidToken)' in err) and self.retries < self.retry:
                # If the credential provider doesn't load new credentials, then it has bad credentials
                # so don't bother trying again.
                if self.credentials_provider.loads_credentials():
                    credentials = self.credentials_provider.credentials()
                    for key, value in credentials.dict().items():
                        kwargs['env'][key] = value

                    self.retries += 1
                    continue

            return 0, output, err



""" Command line interface for the OnePanel Machine Learning platform

'Datasets' commands group.
"""
import os
import re
import threading
import time
import shutil
import logging

import click
import configobj
import humanize
from watchdog.observers import Observer

from onepanel.commands.base import APIViewController
from onepanel.commands.login import login_required
from onepanel.utilities.aws_utility import AWSUtility
from onepanel.utilities.dataset_strings import sanitize_dataset_uid
from onepanel.utilities.s3_authenticator import S3Authenticator
from onepanel.utilities.dataset_downloader import DatasetDownloadListener
from onepanel.utilities.dataset_api import DatasetAPI

from onepanel.utilities.s3.file_sync import FileSynchronizer, ThreadedFileSynchronizer, FileDifference, FileEvent
from onepanel.utilities.s3.file_watch_sync import FileWatchSynchronizerEventHandler
from onepanel.utilities.s3.authentication import APIProvider
from onepanel.utilities.process import run_in_background

import onepanel.utilities.s3.wrapper as S3
import onepanel.services as services


class UploadUpdateThread(threading.Thread):
    def __init__(self, threadID, threadName, vc, awsutil, s3_push_to_dir):
        threading.Thread.__init__(self)
        self.exitFlag = False
        self.threadID = threadID
        self.threadName = threadName
        self.sleepTime = 5  # seconds
        self.vc = vc
        self.awsUtil = awsutil
        self.s3PushToDir = s3_push_to_dir

    def run(self):
        response_data = {}
        while True:
            file_info_s3 = self.awsUtil.get_s3_path_details(self.s3PushToDir)
            update_upload_url = '/update_upload'
            upload_data = {
                'bytesCurrent': int(file_info_s3['data']['total_bytes']),
                'filesCurrent': int(file_info_s3['data']['total_files']),
            }
            response_data['status_code'] = 500
            try:
                response_data = self.vc.post(upload_data, params=update_upload_url)
            except ValueError:
                click.echo("Error with POST, No JSON Object could be decoded.")

            if response_data['status_code'] != 200:
                click.echo(response_data['data'])
                return
            time.sleep(self.sleepTime)
            if self.exitFlag:
                exit(0)


class DatasetViewController(APIViewController):
    """ DatasetViewController data model
    """

    DATASET_FILE = os.path.join('.onepanel', 'dataset')
    EXCLUSIONS = [os.path.join('.onepanel', 'dataset')]

    account_uid = None
    dataset_uid = None

    def __init__(self, conn):
        APIViewController.__init__(self, conn)
        self.endpoint = '{}'.format(
            self.conn.URL,
        )

    def init_credentials_retrieval(self):
        # Figure out the account uid and dataset uid
        home = os.getcwd()
        onepanel_dir = os.path.join(home, '.onepanel')
        if not os.path.exists(onepanel_dir):
            print("ERROR.Directory does not exist, cannot carry out all datasets operations.")
            print("DETAILS." + onepanel_dir)
            exit(-1)
        dataset_file = os.path.join(home, DatasetViewController.DATASET_FILE)
        if not os.path.isfile(dataset_file):
            print("ERROR.Dataset file does not exist, cannot carry out all datasets operations.")
            print("DETAILS." + dataset_file)
            exit(-1)

        cfg = configobj.ConfigObj(dataset_file)

        dataset_uid = cfg['uid']
        dataset_account_uid = cfg['account_uid']

        if len(dataset_uid) < 1 or len(dataset_account_uid) < 1:
            print("ERROR.Dataset file has invalid credentials. Verify credentials or re-pull project.")
            exit(-1)
        self.account_uid = dataset_account_uid
        self.dataset_uid = dataset_uid

    def save(self, home):
        if not os.path.exists(home):
            os.makedirs(home)
        onepanel_dir = os.path.join(home, '.onepanel')
        if not os.path.exists(onepanel_dir):
            os.makedirs(onepanel_dir)
        dataset_file = os.path.join(home, DatasetViewController.DATASET_FILE)

        cfg = configobj.ConfigObj(dataset_file)
        cfg['uid'] = self.dataset_uid
        cfg['account_uid'] = self.account_uid
        cfg.write()

    def init_endpoint(self):
        self.endpoint = '{root}/accounts/{account_uid}/datasets/{dataset_uid}'.format(
            root=self.conn.URL,
            account_uid=self.account_uid,
            dataset_uid=self.dataset_uid
        )

    def mark_version_dirty(self, version, account_uid=None, dataset_uid=None):
        if account_uid is None:
            account_uid = self.account_uid

        if dataset_uid is None:
            dataset_uid = self.dataset_uid

        endpoint = '{root}/accounts/{account_uid}/datasets/{dataset_uid}/version/{version_uid}/mark_dirty'.format(
            root=self.conn.URL,
            account_uid=account_uid,
            dataset_uid=dataset_uid,
            version_uid=version
        )

        return self.put(endpoint=endpoint)

    @classmethod
    def from_json(cls, data):
        cls.account_uid = data['account']['uid']
        cls.dataset_uid = data['uid']

    @classmethod
    def is_uid_valid(cls, uid):
        pattern = re.compile('^[a-z0-9][-a-z0-9]{1,23}[a-z0-9]$')
        if pattern.match(uid):
            return True
        else:
            return False

    @classmethod
    def exists_local(cls, home):
        dataset_file = os.path.join(home, DatasetViewController.DATASET_FILE)
        if os.path.isfile(dataset_file):
            return True
        else:
            return False

    @classmethod
    def exists_remote(cls, dataset_uid, data):
        exists = False
        if data['uid'] == dataset_uid:
            exists = True
        return exists


def check_dataset_exists_remotely(ctx, account_uid, dataset_uid):
    vc = ctx.obj['vc']

    # Check if the dataset already exists remotely
    url_dataset_check = '/accounts/{}/datasets/{}'.format(account_uid, dataset_uid)
    response_data = vc.get(params=url_dataset_check)
    remote_dataset = response_data['data']
    if remote_dataset is not None and vc.exists_remote(dataset_uid, remote_dataset):
        click.echo("Dataset already exists. Please download the dataset if you want to use it.")
        return True
    return False


def init_dataset(ctx, account_uid, dataset_uid, target_dir):
    if not account_uid:
        account_uid = ctx.obj['connection'].account_uid

    if not check_dataset_exists_remotely(ctx, account_uid, dataset_uid):
        create_dataset(ctx, account_uid, dataset_uid, target_dir)
        click.echo('Dataset is initialized in current directory.')


def create_dataset(ctx, account_uid, dataset_uid, target_dir):
    """ Dataset creation method for 'datasets_create' commands
    """
    vc = ctx.obj['vc']

    if not account_uid:
        account_uid = ctx.obj['connection'].account_uid

    can_create = True
    if vc.exists_local(target_dir):
        can_create = click.confirm(
            'Dataset exists locally but does not exist in {}, create the dataset and remap local folder?'
                .format(account_uid))

    if can_create:
        data = {
            'uid': dataset_uid
        }
        url_dataset_create = '/accounts/{}/datasets'.format(account_uid)
        response = vc.post(data, params=url_dataset_create)
        if response['status_code'] == 200:
            vc.from_json(response['data'])
            vc.save(target_dir)
        else:
            click.echo("Encountered error.")
            click.echo(response['status_code'])
            click.echo(response['data'])
            return None
    return target_dir


@click.group(help='Dataset commands group')
@click.pass_context
def datasets(ctx):
    ctx.obj['vc'] = DatasetViewController(ctx.obj['connection'])


@datasets.command('mount-downloader',
                  help='Starts a process in the background that watches for dataset mounts \
                and downloads them as they are created. Resource type is job|instance', hidden=True)
@click.option('-a', '--account_uid', type=str, help='The account uid that the resource is in')
@click.option('-p', '--project_uid', type=str, help='The project uid that the resource is in')
@click.option('-r', '--resource_type', type=str, help='The resource type, e.g. instance|job')
@click.option('-u', '--resource_uid', type=str, help='The resource uid, e.g. instance image-tester')
@click.option('-d', '--download_path', type=str, help='The path where dataset downloads will go to')
@click.option('-r', '--continue_listening', type=bool, help='If true, thread will continue listening for \
                    new dataset mounts. If false, it will stop as soon as there are no more to download.')
@click.option('-v', '--verbose', type=bool, help='If command should log what it is doing')
@click.pass_context
@login_required
def start_dataset_mount_downloader(ctx, account_uid, project_uid, resource_type, resource_uid, download_path, continue_listening, verbose):
    downloader = DatasetDownloadListener(ctx.obj['connection'], download_path, account_uid, project_uid, resource_type, resource_uid, continue_listening, verbose)
    downloader.start()
    downloader.join()


@datasets.command('delete-mount',
                  help='Deletes a dataset mount via API and local files', hidden=True)
@click.option('-a', '--account_uid', type=str, help='The account uid that the resource is in')
@click.option('-p', '--project_uid', type=str, help='The project uid that the resource is in')
@click.option('-d', '--dataset_mount_uuid', type=str, help='The dataset mount uuid')
@click.option('-d', '--download_path', type=str, help='The path where dataset downloads will go to')
@click.option('-v', '--verbose', type=bool, help='If command should log what it is doing')
@click.pass_context
@login_required
def remove_mount(ctx, account_uid, project_uid, dataset_mount_uuid, download_path, verbose):
    if download_path is None:
        download_path = '/onepanel/input/datasets'

    api = DatasetAPI(ctx.obj['connection'])

    logger = logging.getLogger('delete-mount')

    if verbose:
        logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        logger.addHandler(ch)

    try:
        response = api.get_dataset_mount(
            account_uid=account_uid,
            project_uid=project_uid,
            dataset_mount_uuid=dataset_mount_uuid)
    except:
        logger.warning('could not get dataset mount to delete')
        return

    response_code = response['status_code']
    if response_code != 200:
        logger.warning('could not get datasetmount to delete')
        return

    dataset_mount = response['data']
    
    try:
        response = api.delete_dataset_mount(account_uid, project_uid, dataset_mount['uuid'])
    except:
        logger.warning('could not delete dataset mount')
        return

    if not response:
        logger.warning('could not delete dataset mount')
        return

    path = os.path.join(download_path,
                        dataset_mount['dataset']['account']['uid'],
                        dataset_mount['dataset']['uid'],
                        str(dataset_mount['dataset_versioning']['version']))

    # Delete files locally and delete the symlink if any

    if dataset_mount['alias'] is not None:
        path = os.path.join(download_path,
                        '.onepanel',
                        dataset_mount['dataset']['account']['uid'],
                        dataset_mount['dataset']['uid'],
                        str(dataset_mount['dataset_versioning']['version']))

        symlink_path = os.path.join(download_path, dataset_mount['alias'])
    
        if os.path.lexists(symlink_path):
            os.unlink(symlink_path)
            logger.info('deleted symlink')
    else:
        logger.info('No alias to delete')

    logger.info('deleting files under path {}'.format(path))
    shutil.rmtree(path, True)

    logger.info('deleted files')

    logger.info('deleted Dataset Mount - {}'.format(dataset_mount['uuid']))

@datasets.command('list', help='Display a list of all datasets.')
@click.pass_context
@login_required
def datasets_list(ctx):
    vc = ctx.obj['vc']
    data = vc.list(params='/datasets')
    if data is None or len(data['data']) < 1:
        print("No datasets found.")
    else:
        data_print = []
        for datum in data['data']:
            uid = datum['uid']
            if datum['version'] is None:
                version_count = 'No version information provided'
                size = None
            else:
                version_count = datum['version']['version']
                size = datum['version']['size']
            # Some datasets may not have been uploaded to yet
            if size is None:
                size = 0
            size_formatted = humanize.naturalsize(size, binary=True)
            data_print.append({'uid': uid,
                               'version_count': version_count,
                               'size': size_formatted
                               })

        empty_message = "No datasets found."
        fields = ['uid', 'version_count', 'size']
        field_names = ['NAME', 'VERSIONS', 'SIZE']
        DatasetViewController.print_items(data_print, fields, field_names, empty_message)


@datasets.command('init', help='Initialize dataset in current directory.')
@click.pass_context
@login_required
def datasets_init(ctx):
    vc = ctx.obj['vc']
    # Get the parent dir as the default dataset name
    # To support symbolic links as a parent dir name
    wd = os.popen('pwd').readline().strip('\n')
    dataset_uid = os.path.basename(wd)

    # Always prompt user for a dataset name
    suggested = sanitize_dataset_uid(dataset_uid)

    click.echo('Please enter a valid name [{suggested}].'.format(suggested=suggested))
    prompt_msg = click.style('To use suggested name "{suggested}", just press enter'
                             .format(suggested=suggested), fg='blue')
    try:
        dataset_uid = click.prompt(prompt_msg, default="", type=str, show_default=False)
    except click.Abort:
        return None
    if len(dataset_uid) == 0:
        dataset_uid = suggested

    if not vc.is_uid_valid(dataset_uid):
        suggested = sanitize_dataset_uid(dataset_uid)
        prompt_msg = click.style('The name you entered is invalid.',fg='red')
        prompt_msg += click.style('\nName should be 3 to 25 chracters long, lower case alphanumeric or \'-\' '
                      'and must start and end with an alphanumeric character.',fg='red')
        prompt_msg += '\nPlease enter a valid name [{suggested}].'.format(suggested=suggested)
        prompt_msg += click.style('\nTo use suggested name "{alt}", just press enter'.format(alt=suggested),fg='blue')
        try:
            dataset_uid = click.prompt(prompt_msg, default="", type=str, show_default=False)
        except click.Abort:
            return None
        if len(dataset_uid) == 0:
            dataset_uid = suggested
        if not vc.is_uid_valid(dataset_uid):
            click.echo("Invalid dataset name. Please try again.")
            return None
    init_dataset(ctx, None, dataset_uid, os.getcwd())


@datasets.command('create', help='Create dataset in new directory.')
@click.argument('name', type=str)
@click.pass_context
@login_required
def datasets_create(ctx, name):
    vc = ctx.obj['vc']
    target_dir = os.getcwd()
    dataset_uid = name
    if not vc.is_uid_valid(dataset_uid):
        click.echo('Dataset name {} is invalid, please enter a valid name.'.format(dataset_uid))
        click.echo(
            'Name should be 3 to 25 characters long, lower case alphanumeric or \'-\' and must start and end with an alphanumeric character.')
        suggested = dataset_uid.lower()
        suggested = suggested.replace('_', '-')
        prompt_msg = 'Dataset name [{alt}]'.format(alt=suggested)
        try:
            dataset_uid = click.prompt(prompt_msg, default="", type=str, show_default=False)
        except click.Abort:
            return None
        if len(dataset_uid) == 0:
            dataset_uid = suggested
        if not vc.is_uid_valid(dataset_uid):
            click.echo("Dataset name still invalid. Please try again.")
            return None

    account_uid = ctx.obj['connection'].account_uid

    if not check_dataset_exists_remotely(ctx, account_uid, dataset_uid):
        # Attach the desired directory to the current dir
        target_dir += os.sep + dataset_uid
        outcome = create_dataset(ctx, account_uid, dataset_uid, target_dir)
        if outcome is not None:
            click.echo('Dataset is created in directory {}.'.format(outcome))


@datasets.command('push', help='Push up dataset changes')
@click.option(
    '-m', '--message',
    type=str,
    default=None,
    help='Datasets only: Add a message to this version. Up to 255 chars.\"text\".'
)
@click.option(
    '-n', '--name',
    type=str,
    default=None,
    help='Datasets only: Add a name to this version. Use \"text\".'
)
@click.option(
    '-u', '--update-version',
    is_flag=True,
    default=False,
    help='Datasets only, pushes up a new version.'
)
@click.option(
    '-q', '--quiet',
    is_flag=True,
    help='Minimize chatter from executed commands.'
)
@click.option(
    '-b', '--background',
    is_flag=True,
    help='Run the download in the background. Will work even if SSH session is terminated.'
)
@click.option(
    '-w', '--watch',
    is_flag=True,
    help='Datasets only: Syncs up local files to remote. Does not create new dataset versions.'
)
@click.option(
    '-t', '--threads',
    type=int,
    help='Datasets only, when doing a watch: Number of threads allowed.'
)
@click.option(
    '-y', '--yes',
    is_flag=True,
    default=False,
    help='Automatic yes to prompts'
)
@click.pass_context
@login_required
def datasets_push(ctx, message, name, update_version, quiet, background, watch, threads, yes):
    _datasets_push(ctx, message, name, update_version, quiet, background, watch, threads, yes)


def _datasets_push(ctx, message, name, update_version, quiet, background, watch, threads, yes):
    if threads is None:
        threads = 10

    if update_version and background:
        cmd_list = ['onepanel', 'dataset-background-push']
        if message is not None:
            cmd_list.append('--message')
            cmd_list.append('\"' + message + '\"')
        if name is not None:
            cmd_list.append('--name')
            cmd_list.append('\"' + name + '\"')
        if quiet:
            cmd_list.append('-q')
        if background:
            cmd_list.append('-b')

        run_in_background(cmd_list)

        click.echo("Starting upload in the background.")
        return

    if update_version and not background:
        general_push(ctx, message, name, quiet, background)
        return

    if watch and not background:
        click.echo('Starting watch')
        _watch_sync(ctx, threads)
        return

    if watch and background:
        cmd_list = ['onepanel', 'datasets', 'watch', '-t', str(threads)]
        run_in_background(cmd_list)
        click.echo('Starting watch in the background')

        return

    if not update_version and not watch:
        sync(ctx, yes, threads=threads)

@datasets.command('pull', help='Pull down dataset and any changes')
@click.option(
    '-y', '--yes',
    is_flag=True,
    default=False,
    help='Automatic yes to prompts'
)
@click.option(
    '-t', '--threads',
    type=int,
    help='Number of threads allowed for network requests.'
)
@click.pass_context
@login_required
def datasets_pull(ctx, yes, threads):
    _datasets_pull(ctx, yes, threads)


def _datasets_pull(ctx, yes, threads):
    sync(ctx, yes, FileSynchronizer.REMOTE, threads)

@datasets.command(
    'watch',
    hidden=True,
    help='Syncs up local files to remote. Does not create new dataset versions.')
@click.option(
    '-t', '--threads',
    type=int,
    help='Number of threads allowed for network requests.'
)
@click.pass_context
@login_required
def watch(ctx, threads=10):
    _watch_sync(ctx, threads)


def partial_download(ctx, path, destination, version, threads, skip_prompt, delete):
    # Path is {account_uid}/datasets/{dataset_uid}/path/to/folder
    parts = path.split('/', 3)

    if len(parts) < 3:
        click.echo('Path must be account/datasets/dataset/path/to/download')
        return

    account_uid = parts[0]
    dataset_uid = parts[2]
    path = parts[3]

    if destination is None:
        destination = '.'
    download_path = os.path.abspath(os.path.expanduser(destination))

    # +1 for path separator
    cwd_length = len(os.getcwd()) + 1

    ctx.obj['vc'] = DatasetViewController(ctx.obj['connection'])
    vc = ctx.obj['vc']
    vc.account_uid = account_uid
    vc.dataset_uid = dataset_uid
    vc.init_endpoint()

    if threads is None:
        threads = 10

    if version is None:
        version = 'current'

    current_version = vc.get(field_path='/version/' + version)
    if current_version['data'] is None and current_version['status_code'] != 404:
        click.echo('Could not get information about dataset.')
        return

    if current_version['data']['provider']['uid'] != 'aws-s3':
        click.echo('Unsupported dataset provider')
        return

    if path == '/':
        path = ''

    dataset_path = current_version['data']['version']['path'] + '/'

    connection = services.get_connection()
    endpoint = '{}/accounts/{}/datasets/{}/credentials/aws'.format(connection.URL, vc.account_uid, vc.dataset_uid)
    authenticator = APIProvider(connection, endpoint)
    s3_wrapper = S3.Wrapper(credentials_provider=authenticator)

    ignore_file_states = []
    if not delete:
        ignore_file_states.append(FileDifference.State.DELETED)

    master = FileSynchronizer.REMOTE
    synchronizer = FileSynchronizer(download_path, dataset_path, s3_wrapper, master, dataset_path + path, ignore_file_states)
    synchronizer = ThreadedFileSynchronizer(synchronizer, threads, [print_status(cwd_length, master)])

    if not skip_prompt:
        click.echo('Finding file differences...')
    else:
        click.echo('Downloading...')

    differences = synchronizer.find_difference()

    if len(differences) == 0 and not skip_prompt:
        click.echo('No file differences')
        return

    if not skip_prompt:
        for filepath, value in differences.items():
            local_path = filepath[cwd_length:]
            click.echo('{} - {}'.format(local_path, value.state))

        click.confirm('Sync files from remote dataset?', abort=True)

    synchronizer.synchronize(differences.values())
    synchronizer.shutdown()

    click.echo('The files have been downloaded to: {}'.format(download_path))


def datasets_clone(ctx, path, directory, suppress_output=False, run_as_background=False, delete=False):
    conn = ctx.obj['connection']
    vc = DatasetViewController(conn)

    values = path.split('/')

    if len(values) == 3:
        try:
            account_uid, datasets_dir, dataset_uid = values
            assert (datasets_dir == 'datasets')
        except:
            click.echo('Invalid dataset path. Please use <account_uid>/datasets/<uid>')
            return
    else:
        click.echo('Invalid dataset path. Please use <account_uid>/datasets/<uid>')
        return

    vc.account_uid = account_uid
    vc.dataset_uid = dataset_uid

    # check dataset path, account_uid, dataset_uid
    if directory is None:
        home = os.path.join(os.getcwd(), dataset_uid)
    elif directory == '.':
        home = os.getcwd()
    else:
        home = os.path.join(os.getcwd(), directory)

    # check if the dataset exists
    url_dataset_check = '/accounts/{}/datasets/{}'.format(account_uid, dataset_uid)
    response_data = vc.get(uid='', field_path='', params=url_dataset_check)
    response_code = response_data['status_code']
    if response_code == 200:
        remote_dataset = response_data['data']
    elif response_code == 401 or response_code == 404:
        print('Dataset does not exist.')
        return
    else:
        print('Error: {}'.format(response_code))
        return None

    if not vc.exists_remote(dataset_uid, remote_dataset):
        click.echo('There is no dataset {}/datasets/{} on the server'.format(account_uid, dataset_uid))
        return

    can_create = True
    if vc.exists_local(home):
        can_create = click.confirm('Dataset already exists, overwrite?')
    if not can_create:
        return
    # Authenticate for S3
    s3auth = S3Authenticator(ctx.obj['connection'])
    creds = s3auth.get_s3_credentials(vc.account_uid, 'datasets', vc.dataset_uid)
    if creds['data'] is None:
        print("Unable to get S3 credentials. Exiting.")
        return False
    aws_access_key_id = creds['data']['AccessKeyID']
    aws_secret_access_key = creds['data']['SecretAccessKey']
    aws_session_token = creds['data']['SessionToken']
    aws_util = AWSUtility(aws_access_key_id=aws_access_key_id,
                          aws_secret_access_key=aws_secret_access_key,
                          aws_session_token=aws_session_token)
    aws_util.suppress_output = suppress_output
    vc.init_endpoint()

    current_version = vc.get(field_path='/version/current')

    if current_version['data'] is None:
        if current_version['status_code'] == 404:
            click.echo('Dataset is without files.')
            vc.save(home)
            return 0
        else:
            click.echo('Could not get information about dataset.')
            return

    if current_version['data']['provider']['uid'] != 'aws-s3':
        click.echo('Unsupported dataset provider')
        return

    vc.save(home)
    s3_from_dir = current_version['data']['version']['path']
    aws_util.run_cmd_background = run_as_background
    exit_code = aws_util.download_all(home, s3_from_dir, delete)
    if exit_code != 0:
        click.echo('\nError with downloading files.')
        return

    return 0


def general_push(ctx, comment='', name='', suppress_output=False, run_as_background=False):
    num_files = 0
    size_in_bytes_total = 0
    for root, subdirs, files in os.walk(os.getcwd()):
        inside_onepanel_dir = False
        file_path_list = root.split(os.path.sep)
        for path_chunk in file_path_list:
            if '.onepanel' == path_chunk:
                inside_onepanel_dir = True
                break
        if inside_onepanel_dir:
            continue
        for filename in files:
            file_path = os.path.join(root, filename)
            num_files += 1
            size_in_bytes_total += os.path.getsize(file_path)

    if num_files < 1:
        click.echo("Cannot find any files in current dir, exiting.")
        return

    ctx.obj['vc'] = DatasetViewController(ctx.obj['connection'])
    vc = ctx.obj['vc']
    vc.init_credentials_retrieval()
    vc.init_endpoint()

    try:
        # Before trying to do any upload, make sure user has permission
        # For example, we don't want them to modify a public dataset

        url_dataset_member_check = '/is_member'
        response_data = vc.head(uid='', field_path='', params=url_dataset_member_check)

        if response_data['status_code'] == 404:
            click.echo('You are not authorized to push to this dataset.')
            return

        post_obj = {
            'comment': comment,
            'name': name
        }
        update_dataset_resp = vc.put(field_path='/update_version', post_object=post_obj)
        update_dataset_resp_status_code = update_dataset_resp['status_code']
        if (update_dataset_resp_status_code == 201 or update_dataset_resp_status_code == 200) == False:
            click.echo('\nError with calling the API. Contact support if this error continues.')
            return

        if update_dataset_resp_status_code != 201 and update_dataset_resp['data'] is None:
            click.echo('Could not get information about dataset.')
            return

        # Get information about the version, such as the path.
        version_resp = vc.get(field_path='/version/current')
        if version_resp['data'] is None:
            click.echo('Could not get information about dataset.')
            return

        if version_resp['data']['provider']['uid'] != 'aws-s3':
            click.echo('Unsupported dataset provider')
            return

        curr_version = version_resp['data']['version']['version']

        # Start the dataset upload, and check if the dataset is locked from uploading
        data = {
            'bytesTotal': size_in_bytes_total,
            'filesTotal': num_files,
        }
        url_start_upload = '/start_upload'
        try:
            response_data = vc.post(data, params=url_start_upload)
        except ValueError:
            click.echo("Error with POST, No JSON Object could be decoded.")

        if response_data['status_code'] != 200:
            click.echo(response_data['data'])
            return

        # Notify the user, by email, that their background upload is starting
        if run_as_background:
            notify_user_url = '/update_user_for_upload'
            notify_data = {
                'emailType': 'update',
                'emailMsg': 'Expect to see another email once the upload completes.',
            }

            try:
                response_data = vc.post(notify_data, params=notify_user_url)
            except ValueError:
                click.echo("Error with POST, No JSON Object coulzd be decoded.")

            if response_data['status_code'] != 200:
                click.echo(response_data['data'])
                return

        # Authenticate for S3
        s3auth = S3Authenticator(ctx.obj['connection'])
        creds = s3auth.get_s3_credentials(vc.account_uid, 'datasets', vc.dataset_uid)
        if creds['data'] is None:
            print("Unable to get S3 credentials. Exiting.")
            return False
        aws_access_key_id = creds['data']['AccessKeyID']
        aws_secret_access_key = creds['data']['SecretAccessKey']
        aws_session_token = creds['data']['SessionToken']

        dataset_dir = os.curdir
        s3_push_to_dir = version_resp['data']['version']['path']
        aws_util = AWSUtility(aws_access_key_id=aws_access_key_id,
                              aws_secret_access_key=aws_secret_access_key,
                              aws_session_token=aws_session_token)
        aws_util.suppress_output = suppress_output
        aws_util.run_cmd_background = run_as_background
        uploadThread = UploadUpdateThread(1, 'updater', vc, aws_util, s3_push_to_dir)
        uploadThread.start()
        exit_code = aws_util.upload_dir(dataset_dir, s3_push_to_dir, '.onepanel/*')
        if exit_code != 0:
            click.echo('\nError with pushing up files.')
            return

        # Notify the API that the dataset push was completed
        # Need the final dataset information from AWS S3
        file_info_s3 = aws_util.get_s3_path_details(s3_push_to_dir)

        url_finish_upload = '/finish_upload'
        upload_data = {
            'bytesCurrent': int(file_info_s3['data']['total_bytes']),
            'filesCurrent': int(file_info_s3['data']['total_files']),
        }
        try:
            response_data = vc.post(upload_data, params=url_finish_upload)
        except ValueError:
            click.echo("Error with POST, No JSON Object could be decoded.")

        if response_data['status_code'] != 200:
            click.echo(response_data['data'])
            return

        uploadThread.exitFlag = True

        if run_as_background:
            notify_user_url = '/update_user_for_upload'
            notify_data = {
                'emailType': 'update',
                'emailMsg': 'Your dataset has finished uploading.',
            }

            try:
                response_data = vc.post(notify_data, params=notify_user_url)
            except ValueError:
                click.echo("Error with POST, No JSON Object could be decoded.")

            if response_data['status_code'] != 200:
                click.echo(response_data['data'])
                return

        if run_as_background:
            click.echo('\nPushing up version ' + str(curr_version))
        else:
            click.echo('\nPushed up version ' + str(curr_version))
    except Exception as e:
        response_data = {}
        email_msg = [
            'An error occurred during upload.',
            'Details follow',
            e
        ]
        # Notify the user, by email, that their background upload encountered an error
        if True:
            notify_user_url = '/update_user_for_upload'
            notify_data = {
                'emailType': 'error',
                'emailMsg': ' '.join(email_msg),
            }

            try:
                response_data = vc.post(notify_data, params=notify_user_url)
            except ValueError:
                click.echo("Error with POST, No JSON Object coulzd be decoded.")

            if response_data['status_code'] != 200:
                click.echo(response_data['data'])
                return


def datasets_download(ctx, path, directory, suppress_output=False, run_as_background=False, delete=False):
    if directory is None or directory == '.':
        home = os.getcwd()
    else:
        home = os.path.join(os.getcwd(), directory)

    code = datasets_clone(ctx, path, home, suppress_output=suppress_output, run_as_background=run_as_background, delete=delete)
    if code != 0:
        print("Unable to download!")
        return False

    if run_as_background is False:
        print('The files have been downloaded to: {dir}'.format(dir=home))
    return True


def sync(ctx, skip_check, master=FileSynchronizer.LOCAL, threads=10):
    # +1 for path separator
    cwd_length = len(os.getcwd()) + 1

    ctx.obj['vc'] = DatasetViewController(ctx.obj['connection'])
    vc = ctx.obj['vc']
    vc.init_credentials_retrieval()
    vc.init_endpoint()

    if threads is None:
        threads = 10

    current_version = vc.get(field_path='/version/current')
    if current_version['data'] is None and current_version['status_code'] != 404:
        click.echo('Could not get information about dataset.')
        return

    if current_version['data']['provider']['uid'] != 'aws-s3':
        click.echo('Unsupported dataset provider')
        return

    dataset_path = current_version['data']['version']['path'] + '/'

    connection = services.get_connection()
    endpoint = '{}/accounts/{}/datasets/{}/credentials/aws'.format(connection.URL, vc.account_uid, vc.dataset_uid)
    authenticator = APIProvider(connection, endpoint)
    s3_wrapper = S3.Wrapper(credentials_provider=authenticator)

    synchronizer = FileSynchronizer(os.getcwd(), dataset_path, s3_wrapper, master)
    synchronizer = ThreadedFileSynchronizer(synchronizer, threads, [print_status(cwd_length, master)])

    click.echo('Finding file differences...')

    differences = synchronizer.find_difference()

    if len(differences) == 0:
        click.echo('No file differences')
        return

    if not skip_check:
        for filepath, value in differences.items():
            local_path = filepath[cwd_length:]
            click.echo('{} - {}'.format(local_path, value.state))

        if master == FileSynchronizer.LOCAL:
            click.confirm('Sync files to remote dataset?', abort=True)
        else:
            click.confirm('Sync files from remote dataset?', abort=True)

    synchronizer.synchronize(differences.values())
    synchronizer.shutdown()

    vc.mark_version_dirty(current_version['data']['version']['version'])


def _watch_sync(ctx, threads=10):
    """Watches the current directory dataset for changes, and syncs them up to remote dataset"""
    # +1 for path separator
    cwd_length = len(os.getcwd()) + 1
    cwd = os.getcwd()

    ctx.obj['vc'] = DatasetViewController(ctx.obj['connection'])
    vc = ctx.obj['vc']
    vc.init_credentials_retrieval()
    vc.init_endpoint()

    current_version = vc.get(field_path='/version/current')
    if current_version['data'] is None and current_version['status_code'] != 404:
        click.echo('Could not get information about dataset.')
        return

    if current_version['data']['provider']['uid'] != 'aws-s3':
        click.echo('Unsupported dataset provider')
        return

    dataset_path = current_version['data']['version']['path'] + '/'

    connection = services.get_connection()
    endpoint = '{}/accounts/{}/datasets/{}/credentials/aws'.format(connection.URL, vc.account_uid, vc.dataset_uid)
    authenticator = APIProvider(connection, endpoint)
    s3_wrapper = S3.Wrapper(credentials_provider=authenticator)
    synchronizer = FileSynchronizer(os.getcwd(), dataset_path, s3_wrapper)

    def update_api_on_success(file_event):
        if file_event.state != FileEvent.FAILED:
            vc.mark_version_dirty(current_version['data']['version']['version'])

    synchronizer = ThreadedFileSynchronizer(synchronizer, threads, [print_status(cwd_length), update_api_on_success])

    observer = Observer()
    observer.schedule(FileWatchSynchronizerEventHandler(cwd, dataset_path, synchronizer, ignore_patterns=['.*', '*.swx', '*.swpx', '*.swp', '*~', '*.sb-']),
                      path=cwd, recursive=True)
    observer.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()

    observer.join()


def print_status(cwd_length, master=FileSynchronizer.LOCAL):
    """Returns a function that takes in a file_difference and prints the action on the input file
    Uses the current working directory length to print relative filenames.
    """

    def print_status_internal_local(file_event):
        """

         :param file_event:
         :type file_event FileEvent
         :return:
         """
        file_difference = file_event.file_difference

        local_path = file_difference.source_path[cwd_length:]

        if file_difference.state == FileDifference.State.MOVED:
            local_source_path = file_difference.original_source_path[cwd_length:]

            if file_event.state == FileEvent.FINISHED:
                click.echo('Moved {} to {}'.format(local_source_path, local_path))
            elif file_event.state == FileEvent.FAILED:
                click.echo('(FAILED) Moving {} to {}'.format(local_source_path, local_path))
            else:
                click.echo('Moving {} to {}'.format(local_source_path, local_path))

            return

        action = 'Uploading'

        if file_difference.state == FileDifference.State.DELETED:
            action = 'Deleting'

        if file_difference.state == FileDifference.State.MODIFIED:
            action = 'Uploading Modified'

        if file_event.state == FileEvent.FINISHED:
            action = action.replace('ing', 'ed')
        elif file_event.state == FileEvent.FAILED:
            action = '(FAILED) {} - {}'.format(action, file_event.result)

        click.echo('{} - {}'.format(action, local_path))

    def print_status_internal_remote(file_event):
        """

        :param file_event:
        :type file_event FileEvent
        :return:
        """
        file_difference = file_event.file_difference

        local_path = file_difference.source_path[cwd_length:]

        if file_difference.state == FileDifference.State.MOVED:
            local_source_path = file_difference.original_source_path[cwd_length:]
            click.echo('Moving {} to {}'.format(local_source_path, local_path))
            return

        action = 'Downloading'

        if file_difference.state == FileDifference.State.DELETED:
            action = 'Deleting'

        if file_difference.state == FileDifference.State.MODIFIED:
            action = 'Downloading Modified'

        if file_event.state == FileEvent.FINISHED:
            action = action.replace('ing', 'ed')
        elif file_event.state == FileEvent.FAILED:
            action = '(FAILED) {} - {}'.format(action, file_event.result)

        click.echo('{} - {}'.format(action, local_path))

    if master == FileSynchronizer.LOCAL:
        return print_status_internal_local

    return print_status_internal_remote



from setuptools import setup

from onepanel.constants import *

setup(
    name="onepanel",
    version=CLI_VERSION,
    packages=['onepanel', 'onepanel.commands', 'onepanel.types', 'onepanel.models',
              'onepanel.utilities', 'onepanel.utilities.s3', 'onepanel.git_hooks'],
    python_requires='>=2.6, !=3.0.*, !=3.1.*, !=3.2.*, <4',
    install_requires=[
        'configparser',
        'PyYAML<=3.13,>=3.10',
        'prettytable',
        'requests',
        'click>=7',
        'PTable',
        'configobj',
        'websocket-client',
        'humanize',
        'botocore>=1.12.0,<1.13.0'
        'awscli>=1.16.0,<1.17.0',
        'boto3>=1.9.0,<1.10.0',
        'watchdog',
        'iso8601',
        'future',
        'psutil',
        'nvidia-ml-py3'
    ],
    setup_requires=[
    ],
    entry_points='''
        [console_scripts]
        onepanel=onepanel.cli:main
    ''',
)



""" Command line interface for the OnePanel Machine Learning platform

'ProjectViewControllers' commands group.
"""

import os
import re

import click
import configobj

from onepanel.commands.base import APIViewController
from onepanel.commands.login import login_required
from onepanel.gitwrapper import GitWrapper


class ProjectViewController(APIViewController):
    """ ProjectViewControllers data model
    """

    PROJECT_FILE = os.path.join('.onepanel','project')
    EXCLUSIONS = [os.path.join('.onepanel','project')]

    account_uid = None
    project_uid = None

    def __init__(self, conn):
        APIViewController.__init__(self, conn)


    def init_credentials_retrieval(self):
        home = os.getcwd()
        onepanel_dir = os.path.join(home, '.onepanel')
        if not os.path.exists(onepanel_dir):
            print("ERROR.Directory does not exist, cannot carry out all projects operations.")
            print("DETAILS." + onepanel_dir)
            exit(-1)
        project_file = os.path.join(home, ProjectViewController.PROJECT_FILE)
        if not os.path.isfile(project_file):
            print("ERROR.Project file does not exist, cannot carry out all projects operations.")
            print("DETAILS." + project_file)
            exit(-1)

        cfg = configobj.ConfigObj(project_file)
        project_account_uid = cfg['account_uid']
        project_uid = cfg['uid']

        if len(project_account_uid) < 1:
            print("ERROR.Project file has invalid credentials. Verify credentials or re-pull project.")
            exit(-1)
        self.account_uid = project_account_uid
        self.project_uid = project_uid

    def save(self, home):
        if not os.path.exists(home):
            os.makedirs(home)
        onepanel_dir = os.path.join(home, '.onepanel')
        if not os.path.exists(onepanel_dir):
            os.makedirs(onepanel_dir)
        project_file = os.path.join(home, ProjectViewController.PROJECT_FILE)

        cfg = configobj.ConfigObj(project_file)
        cfg['uid'] = self.project_uid
        cfg['account_uid'] = self.account_uid
        cfg.write()

    @classmethod
    def from_json(cls, data):
        cls.account_uid = data['account']['uid']
        cls.project_uid = data['uid']

    @classmethod
    def from_directory(cls, home):
        if not cls.exists_local(home):
            print("ERROR.Cannot find path.")
            print("DETAILS."+home)
            exit(-1)

        project_file = os.path.join(home, ProjectViewController.PROJECT_FILE)
        cfg = configobj.ConfigObj(project_file)
        cls.account_uid = cfg['account_uid']
        cls.project_uid = cfg['uid']

    @classmethod
    def is_uid_valid(cls,uid):
        pattern = re.compile('^[a-z0-9][-a-z0-9]{1,23}[a-z0-9]$')
        if pattern.match(uid):
            return True
        else:
            return False

    @classmethod
    def exists_local(cls,home):
        project_file = os.path.join(home, cls.PROJECT_FILE)
        if os.path.isfile(project_file):
            return True
        else:
            return False

    @classmethod
    def exists_remote(cls,project_uid, data):
        exists = False
        if data is None:
            return exists
        if data['uid'] == project_uid:
            exists = True
        return exists


def create_project(ctx, account_uid, home):
    """ Project creation method for 'projects_init' and 'projects_create'
    commands
    """
    vc = ctx.obj['vc']

    if not account_uid:
        account_uid = ctx.obj['connection'].account_uid

    project_uid = os.path.basename(home)
    if not vc.is_uid_valid(project_uid):
        click.echo('Project name {} is invalid.'.format(project_uid))
        click.echo(
            'Name should be 3 to 25 characters long, lower case alphanumeric or \'-\' and must start and end with an alphanumeric character.')
        return None

    url_project_check = '/accounts/{}/projects/{}'.format(account_uid, project_uid)
    response_data = vc.get(params=url_project_check,uid='',field_path='')
    remote_project = response_data['data']

    project = None
    if vc.exists_remote(project_uid, remote_project):
        if vc.exists_local(home):
            click.echo('Project is already initialized')
        else:
            vc.account_uid = account_uid
            vc.project_uid = project_uid
            vc.save(home)
            git = GitWrapper()
            git.init(onepanel_user_uid=ctx.obj['connection'].user_uid, gitlab_token=vc.conn.gitlab_impersonation_token,
                     home=home, account_uid=account_uid, project_uid=project_uid)
            git.exclude(home, ProjectViewController.EXCLUSIONS)

    else:
        can_create = True
        if vc.exists_local(home):
            can_create = click.confirm(
                'Project exists locally but does not exist in {}, create the project and remap local folder?'
                    .format(account_uid))

        if can_create:
            data = {
                'uid': project_uid
            }
            url_project_create = '/accounts/{}/projects'.format(account_uid)
            response = vc.post(data,params=url_project_create)
            if response['status_code'] == 200:
                vc.from_json(response['data'])
                vc.save(home)
                git = GitWrapper()
                git.init(onepanel_user_uid=ctx.obj['connection'].user_uid, gitlab_token=vc.conn.gitlab_impersonation_token,
                         home=home, account_uid=account_uid, project_uid=project_uid)
                git.exclude(home, ProjectViewController.EXCLUSIONS)
    return project


@click.group(help='Project commands group')
@click.pass_context
def projects(ctx):
    ctx.obj['vc'] = ProjectViewController(ctx.obj['connection'])


@projects.command('list', help='Display a list of all projects.')
@click.pass_context
@login_required
def projects_list(ctx):
    vc = ctx.obj['vc']
    data = vc.list(params='/projects')
    empty_message = "No projects found"
    items = []
    if data is not None:
        items = data['data']
    vc.print_items(
        items,
        fields=['uid', 'instanceCount', 'jobCount'],
        field_names=['NAME', 'INSTANCES', 'JOBS'],
        empty_message=empty_message
    )

@projects.command('init', help='Initialize project in current directory.')
@click.pass_context
@login_required
def projects_init(ctx):
    home = os.getcwd()
    vc = ctx.obj['vc']
    if not vc.is_uid_valid(os.path.basename(home)):
        project_uid = click.prompt('Please enter a valid project name')
        home = os.path.join(home, project_uid)

    if create_project(ctx, None, home):
        click.echo('Project is initialized in current directory.')


@projects.command('create', help='Create project in new directory.')
@click.argument('name', type=str)
@click.pass_context
@login_required
def projects_create(ctx, name):
    home = os.path.join(os.getcwd(), name)
    if create_project(ctx, None, home):
        click.echo('Project is created in directory {}.'.format(home))


def projects_clone(ctx, path, directory):
    conn = ctx.obj['connection']
    vc = ProjectViewController(conn)

    values = path.split('/')

    if len(values) == 3:
        try:
            account_uid, projects_dir, project_uid = values
            assert (projects_dir == 'projects')
        except:
            click.echo('Invalid project path. Please use <account_uid>/projects/<uid>')
            return
    else:
        click.echo('Invalid project path. Please use <account_uid>/projects/<uid>')
        return

    # check project path, account_uid, project_uid
    if directory is None:
        home = os.path.join(os.getcwd(), project_uid)
    elif directory == '.':
        home = os.getcwd()
    else:
        home = os.path.join(os.getcwd(), directory)

    # check if the project exists
    url_project_check = '/accounts/{}/projects/{}'.format(account_uid, project_uid)
    response_data = vc.get(params=url_project_check,uid="",field_path='')
    response_code = response_data['status_code']
    if response_code == 200:
        remote_project = response_data['data']
    elif response_code == 401 or response_code == 404:
        print('Project does not exist.')
        return
    else:
        print('Error: {}'.format(response_code))
        return

    if not vc.exists_remote(project_uid, remote_project):
        click.echo('There is no project {}/projects/{} on the server'.format(account_uid, project_uid))
        return

    can_create = True
    if vc.exists_local(home):
        can_create = click.confirm('Project already exists, overwrite?')
    if not can_create:
        return

    # git clone
    git = GitWrapper()
    git_clone_code = git.clone(conn.user_uid,conn.gitlab_impersonation_token,home, account_uid, project_uid)
    if git_clone_code == 0:
        vc.account_uid = account_uid
        vc.project_uid = project_uid
        vc.save(home)
        git.exclude(home, ProjectViewController.EXCLUSIONS)
        git.git_hook_code(home)
    elif git_clone_code == 128:
        print('ERROR. Git reports the Path already exists and is not an empty directory.')
    else:
        print('ERROR. Git reported a problem with cloning your project.')


import re
import datetime

from onepanel.models.api_json import APIJSON
from onepanel.utilities.time import parse_date, format_timedelta, utcnow
from onepanel.models.machine_type import MachineType
from onepanel.models.volume_type import VolumeType
from onepanel.models.project import Project
from onepanel.models.instance_template import InstanceTemplate
from onepanel.models.dataset_mount_identifier import DatasetMountIdentifier
from onepanel.models.git_source import GitSource


class Instance(APIJSON):
    CREATED = 'created'
    STARTING = 'starting'
    STARTED = 'started'
    READY = 'ready'
    PAUSED = 'paused'
    RESUMING = 'resuming'
    RESUMED = 'resumed'

    def __init__(self):
        self.uid = None
        self.created_at = None
        self.launched_at = None
        self.started_at = None
        self.ready_at = None
        self.last_stopped_at = None
        self.active = False
        self.project = Project()
        self.machine_type = MachineType()
        self.volume_type = VolumeType()
        self.instance_template = InstanceTemplate()
        self.dataset_mounts = []
        self.dataset_mount_claims = []
        self.git_source = None

    @classmethod
    def from_json(cls, dct):
        instance = cls()

        instance.set_uid(dct['uid'])
        instance.created_at = parse_date(dct['createdAt'])

        #LaunchedAt is ommited from API if it is empty 
        if 'launchedAt' in dct:
            instance.launched_at = parse_date(dct['launchedAt'])
        
        instance.started_at = parse_date(dct['startedAt'])
        instance.ready_at = parse_date(dct['readyAt'])
        instance.last_stopped_at = parse_date(dct['lastStoppedAt'])
        instance.active = dct['active']
        instance.project = Project.from_json(dct['project'])
        instance.machine_type = MachineType.from_json(dct['machineType'])
        instance.volume_type = VolumeType.from_json(dct['volumeType'])
        instance.instance_template = InstanceTemplate.from_json(dct['instanceTemplate'])
        instance.git_source = GitSource.from_json(dct['gitSource'])

        return instance

    @classmethod
    def from_simple_json(cls, dct):
        instance = cls()

        instance.set_uid(dct['name'])
        instance.machine_type.uid = dct['machine']
        instance.volume_type.uid = dct['storage']
        instance.instance_template.uid = dct['environment']
        instance.dataset_mount_identifiers = [DatasetMountIdentifier.from_simple_json(item) for item in dct['datasets']]
        instance.git_source = GitSource.from_string(dct['git'])

        return instance

    def set_uid(self, uid):
        pattern = re.compile('^[a-z0-9][-a-z0-9]{1,23}[a-z0-9]$')

        if not pattern.match(uid):
            raise ValueError(
                'Name should be 3 to 25 characters long, lower case alphanumeric or \'-\' and must start and end with an alphanumeric character.')

        self.uid = uid

    # TODO document
    def simple_view(self):
        info = self.machine_type.info

        item = {
            'account_uid': self.project.account.uid,
            'project_uid': self.project.uid,
            'uid': self.uid,
            'cpu': info['cpu'],
            'gpu': info['gpu'],
            'ram': info['ram'],
            'hdd': self.volume_type.info['size'],
            'duration_ready': format_timedelta(self.duration_ready()),
            'duration_paused': format_timedelta(self.duration_paused()),
            'age': format_timedelta(self.age()),
            'state': self.state()
        }

        return item

    def state(self):
        if not self.active:
            if self.last_stopped_at is not None:
                return Instance.PAUSED

            return Instance.CREATED

        if self.ready_at is not None:
            return Instance.READY

        if self.started_at is not None:
            if self.last_stopped_at is not None:
                return Instance.RESUMED

            return Instance.STARTED

        if self.last_stopped_at is not None:
            return Instance.RESUMING

        return Instance.STARTING

    def duration_ready(self, when=None):
        """How long the instance has been ready for. This is calculated from the last time it was 
           marked ready to when."""
        if when is None:
            when = utcnow()

        if self.state() != Instance.READY:
            return None

        return when - self.ready_at

    def duration_paused(self, when=None):
        """How long the instance has been paused for, if it is paused. If not paused, returns None"""
        if self.state() != Instance.PAUSED:
            return None

        if when is None:
            when = utcnow()

        return when - self.last_stopped_at

    def age(self, when=None):
        """How long the instance has been created to when"""
        if when is None:
            when = utcnow()

        if self.created_at is None:
            return datetime.timedelta(0)

        return when - self.created_at

    def api_json(self):
        json = {
            'uid': self.uid,
            'machineType': self.machine_type,
            'volumeType': self.volume_type,
            'instanceTemplate': self.instance_template,
            'datasetMountClaims': self.dataset_mount_claims,
        }

        if self.git_source is not None:
            json['gitSource'] = self.git_source

        return json



import os

import boto3
from boto3.exceptions import S3UploadFailedError
from botocore.exceptions import ClientError

from onepanel.utilities.s3.authentication import Provider


class Wrapper:
    @staticmethod
    def is_error_expired_token(error):
        """
        :param error:
        :type error ClientError
        :return:
        """

        return error.response['Error']['Code'] == 'ExpiredToken'

    def __init__(self, bucket_name=None, credentials_provider=None, retry=3):
        """
        :param bucket_name: AWS S3 Bucket name
        :param credentials_provider: Provides credentials for requests.
        :type credentials_provider: onepanel.lib.s3.authentication.Provider
        """
        if bucket_name is None:
            bucket_name = os.getenv('DATASET_BUCKET', 'onepanel-datasets')

        if credentials_provider is None:
            credentials_provider = Provider()

        self.bucket_name = bucket_name
        self.credentials_provider = credentials_provider
        self.s3 = None
        self.reset_client()
        self.retry = retry
        self.retries = 0

    def create_client(self):
        if not self.credentials_provider.loads_credentials():
            return boto3.client('s3')

        credentials = self.credentials_provider.credentials()

        return boto3.client('s3',
                            aws_access_key_id=credentials.access_key_id,
                            aws_secret_access_key=credentials.secret_access_key,
                            aws_session_token=credentials.session_token)

    def create_resource(self):
        if not self.credentials_provider.loads_credentials():
            return boto3.resource('s3')

        credentials = self.credentials_provider.credentials()

        return boto3.resource('s3',
                            aws_access_key_id=credentials.access_key_id,
                            aws_secret_access_key=credentials.secret_access_key,
                            aws_session_token=credentials.session_token)

    def reset_client(self):
        self.s3 = self.create_client()

    def list_files(self, prefix):
        paginator = self.s3.get_paginator('list_objects_v2')

        has_more = True
        next_marker = ''

        files = {}

        while has_more:
            response_iterator = paginator.paginate(
                Bucket=self.bucket_name,
                Prefix=prefix,
                StartAfter=next_marker,
                PaginationConfig={
                    'MaxItems': 1000,
                    'PageSize': 1000
                }
            )

            content_length = 0

            try:
                for page in response_iterator:
                    if 'Contents' in page:
                        for content in page['Contents']:
                            key = content['Key']
                            files[key] = content
                            next_marker = key
                            content_length += 1
            except ClientError as clientError:
                if Wrapper.is_error_expired_token(clientError) and self.retries < self.retry:
                    self.retries += 1
                    self.reset_client()
                    return self.list_files(prefix)

                self.retries = 0
                raise

            if content_length == 0:
                has_more = False
                break

        # Skip the file that has the same name as the prefix.
        if prefix in files:
            del files[prefix]

        self.retries = 0
        return files

    def upload_file(self, filepath, key):
        # Normalize s3 path if we're on windows
        key = key.replace('\\', '/')

        try:
            result = self.s3.upload_file(filepath, self.bucket_name, key)
            self.retries = 0
            return result
        except ClientError as clientError:
            if Wrapper.is_error_expired_token(clientError) and self.retries < self.retry:
                self.retries += 1
                self.reset_client()
                return self.upload_file(filepath, key)

            raise
        except S3UploadFailedError as uploadError:
            if self.retries < self.retry:
                self.retries += 1
                self.reset_client()
                return self.upload_file(filepath, key)

            raise
        except FileNotFoundError as fileNotFoundError:
            # Do nothing
            return

    def download_file(self, local_path, key):
        s3 = self.create_resource()

        try:
            # Create the directory locally if it doesn't exist yet.
            dirname = os.path.dirname(local_path)
            os.makedirs(dirname, exist_ok=True)
            s3.Bucket(self.bucket_name).download_file(key, local_path)
        except ClientError as clientError:
            if self.retries < self.retry:
                self.retries += 1
                self.reset_client()
                return self.download_file(local_path, key)

            raise

    def delete_file(self, key):
        key = key.replace('\\', '/')

        try:
            result = self.s3.delete_object(
                Bucket=self.bucket_name,
                Key=key
            )

            self.retries = 0
            return result
        except ClientError as clientError:
            if Wrapper.is_error_expired_token(clientError) and self.retries < self.retry:
                self.retries += 1
                self.reset_client()
                return self.delete_file(key)

            raise

    def copy_file(self, source_key, destination_key):
        # Normalize s3 path if we're on windows
        source_key = source_key.replace('\\', '/')
        destination_key = destination_key.replace('\\', '/')

        s3 = self.create_resource()
        copy_source = {
            'Bucket': self.bucket_name,
            'Key': source_key
        }

        try:
            s3.meta.client.copy(copy_source, self.bucket_name, destination_key)
        except ClientError as clientError:
            if self.retries < self.retry:
                self.retries += 1
                self.reset_client()
                return self.copy_file(source_key, destination_key)

            raise

    def move_file(self, source_key, destination_key):
        # We do a try here in case copy fails, in which case we don't try delete.
        try:
            self.copy_file(source_key, destination_key)
            self.delete_file(source_key)
        except BaseException:
            raise



import onepanel.git_hooks.pre_commit_content


def read_self():
	file_path = __file__
	if file_path.endswith('.pyc'):
		file_path = file_path.replace('pyc','py')
	with open(file_path,'r') as f:
		return f.read()


if __name__ == '__main__':
	onepanel.git_hooks.pre_commit_content.check_for_large_files()



import os


def get_file_tree(path, skip_hidden=True):
    """
    Finds all files in the specified path and subdirectories.

    :param path:
    :type path str
    :return: a map with key being the file path and value being the statistics of the file
    """

    result = []

    for root, subdirs, files in os.walk(path):
        if skip_hidden:
            files = [f for f in files if not f[0] == '.']
            subdirs[:] = [d for d in subdirs if not d[0] == '.']

        for name in files:
            result.append(os.path.join(root, name))

    return result



import os
import subprocess
import sys
import logging

import click

from onepanel.commands import jobs
from onepanel.commands.datasets import datasets_clone, datasets_download, DatasetViewController, general_push, _datasets_pull, _datasets_push, partial_download
from onepanel.commands.jobs import jobs_download_output, JobViewController
from onepanel.commands.login import login_required
from onepanel.commands.projects import projects_clone, ProjectViewController
from onepanel.gitwrapper import GitWrapper
from onepanel.utilities.timer import Timer

COMMON_LOGGER = 'common_logger'


def get_entity_type(path):
    dirs = path.split('/')

    entity_type = None
    if len(dirs) == 3:
        account_uid, dir, uid = dirs
        if dir == 'projects':
            entity_type = 'project'
        elif dir == 'datasets':
            entity_type = 'dataset'
    elif len(dirs) == 5:
        account_uid, parent, project_uid, dir, uid = dirs
        if parent == 'projects' and dir == 'jobs':
            entity_type = 'job'

    if len(dirs) > 2 and dirs[1] == 'datasets':
        entity_type = 'dataset'

    return entity_type

@click.command('clone', help='Clone project or dataset from server.')
@click.argument('path', type=click.Path())
@click.argument('directory', type=click.Path(), required=False)
@click.option(
    '-q', '--quiet',
    is_flag=True,
    help='Minimize chatter from executed commands.'
)
@click.pass_context
@login_required
def clone(ctx, path, directory,quiet):
    entity_type = get_entity_type(path)
    if entity_type == 'project':
        projects_clone(ctx, path, directory)
    elif entity_type == 'dataset':
        datasets_clone(ctx, path, directory,quiet)
    else:
        click.echo('Invalid project or dataset path.')

@click.command('download', help='Download a dataset')
@click.argument('path', type=click.Path())
@click.argument('directory', type=click.Path(), required=False)
@click.option(
    '--delete',
    type=bool,
    is_flag=True,
    default=False,
    help='Deletes files locally that are not in the dataset'
)
@click.option(
    '--archive',
    type=bool,
    is_flag=True,
    default=False,
    help='Download the output as a compressed file.'
)
@click.option(
    '-q', '--quiet',
    is_flag=True,
    help='Minimize chatter from executed commands.'
)
@click.option(
    '-b', '--background',
    is_flag=True,
    help='Run the download in the background. Will work even if SSH session is terminated.'
)
@click.option(
    '-v', '--version',
    type=str,
    help='The version of the dataset. If none is provided, latest is used.')
@click.option(
    '-t', '--threads',
    type=int,
    help='Number of threads allowed.'
)
@click.pass_context
@login_required
def download(ctx, path, directory, archive, quiet, background, version, threads, delete):
    entity_type = get_entity_type(path)
    if entity_type == 'dataset':
        if path.count('/') == 2:
            datasets_download(ctx, path, directory, quiet, background, delete)
        else:
            partial_download(ctx, path, directory, version, threads, True, delete)
    elif entity_type == 'job':
        jobs_download_output(ctx, path, directory, archive)
    else:
        click.echo('Invalid path.')


@click.command('push', help='Push changes to onepanel')
@click.option(
    '-m', '--message',
    type=str,
    default=None,
    help='Datasets only: Add a message to this version. Up to 255 chars.\"text\".'
)
@click.option(
    '-n', '--name',
    type=str,
    default=None,
    help='Datasets only: Add a name to this version. Use \"text\".'
)
@click.option(
    '-u', '--update-version',
    is_flag=True,
    default=False,
    help='Datasets only, pushes up a new version.'
)
@click.option(
    '-q', '--quiet',
    is_flag=True,
    help='Minimize chatter from executed commands.'
)
@click.option(
    '-b', '--background',
    is_flag=True,
    help='Run the download in the background. Will work even if SSH session is terminated.'
)
@click.option(
    '-w', '--watch',
    is_flag=True,
    help='Datasets only: Syncs up local files to remote. Does not create new dataset versions.'
)
@click.option(
    '-t', '--threads',
    type=int,
    help='Datasets only, when doing a watch: Number of threads allowed.'
)
@click.option(
    '-y', '--yes',
    is_flag=True,
    default=False,
    help='Automatic yes to prompts'
)
@click.pass_context
@login_required
def push(ctx, message, name, quiet, background, update_version, watch, threads, yes):
    home = os.getcwd()
    # Are we uploading job output? A project? Or Dataset?
    if os.path.isfile(JobViewController.JOB_OUTPUT_FILE):
        jobs.upload_output(ctx, quiet)
    elif os.path.isfile(ProjectViewController.PROJECT_FILE):
        if message is not None or name is not None:
            click.echo(
                "Projects do not support these arguments. Remove arguments and try again.")
            return
        GitWrapper().push(home)
    elif os.path.isfile(DatasetViewController.DATASET_FILE):
        if background and update_version:
            close_fds = False
            cmd_list = ['onepanel', 'dataset-background-push']
            if message is not None:
                cmd_list.append('--message')
                cmd_list.append('\"'+message+'\"')
            if name is not None:
                cmd_list.append('--name')
                cmd_list.append('\"'+name+'\"')
            if quiet:
                cmd_list.append('-q')
            if background:
                cmd_list.append('-b')
            if sys.platform != 'win32':
                cmd_list.insert(0, 'nice')
                cmd_list.insert(0, 'nohup')
                close_fds = True
            else:
                # /i so that windows doesn't create "%SYSTEM_DRIVE%" folder
                cmd_list.insert(0, 'start /b /i')
            cmd = ' '.join(cmd_list)
            if sys.platform != 'win32':
                stdout = open(os.path.devnull,'a')
                stderr = open(os.path.devnull,'a')
                subprocess.Popen(args=cmd, stdin=subprocess.PIPE, stdout=stdout,
                                 stderr=stderr, shell=True, close_fds=close_fds,preexec_fn=os.setpgrp)
            else:
                # Windows specific instructions
                # https://docs.microsoft.com/en-us/windows/desktop/ProcThread/process-creation-flags
                CREATE_NO_WINDOW = 0x08000000
                CREATE_NEW_PROCESS_GROUP = 0x00000200
                subprocess.Popen(args=cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE, shell=True, close_fds=close_fds,
                                 creationflags=CREATE_NO_WINDOW | CREATE_NEW_PROCESS_GROUP)

            click.echo("Starting upload in the background.")
        elif update_version:
            general_push(ctx, message, name, quiet, background)
        else:
            _datasets_push(ctx, message, name, update_version, quiet, background, watch, threads, yes)
    else:
        click.echo("Cannot determine if you are trying to push job output, dataset files, or project files. Are you in the right directory?")

@click.command('dataset-background-push', help='This is to be called by a python sub-process.',hidden=True)
@click.option(
    '-m', '--message',
    type=str,
    default=None,
    help='Datasets only: Add a message to this version. Up to 255 chars.\"text\".'
)
@click.option(
    '-n', '--name',
    type=str,
    default=None,
    help='Datasets only: Add a name to this version. Use \"text\".'
)
@click.option(
    '-q', '--quiet',
    is_flag=True,
    help='Minimize chatter from executed commands.'
)
@click.option(
    '-b','--background',
    is_flag=True,
    help='Run the download in the background. Will work even if SSH session is terminated.'
)
@click.pass_context
@login_required
def background_dataset_push(ctx,message,name,quiet,background):
    general_push(ctx, message, name, quiet, background)


@click.command('pull', help='Pull changes from onepanel (fetch and merge)')
@click.option(
    '-y', '--yes',
    is_flag=True,
    default=False,
    help='Automatic yes to prompts'
)
@click.option(
    '-t', '--threads',
    type=int,
    help='Number of threads allowed for network requests.'
)
@click.pass_context
@login_required
def pull(ctx, yes, threads):
    if os.path.isfile(ProjectViewController.PROJECT_FILE):
        home = os.getcwd()
        GitWrapper().pull(home)
    elif os.path.isfile(DatasetViewController.DATASET_FILE):
        _datasets_pull(ctx, yes, threads)
    else:
        click.echo('Path is not a project or dataset. Nothing to pull.')


@click.command('timer-sync-output', help='Syncs job output and logs repeated on a timer', hidden=True)
@click.option('-j', '--job_uid', type=str, help='The job uid')
@click.option('-p', '--project_uid', type=str, help='The uid of the project the job is for')
@click.option('-a', '--project_account_uid', type=str, help='The uid of the account that owns the project the job is in')
@click.option('-d', '--delay', type=str, help='What interval the timer sends a request')
@click.option('-d', '--path', type=str, help='Path to sync files from')
@click.option('-d', '--destination', type=str, help='Path to sync files to')
@click.option('-v', '--verbose', type=bool, help='If command should log what it is doing')
@click.pass_context
@login_required
def timer_sync_output(ctx, job_uid, project_uid, project_account_uid, path, destination, delay, verbose):
    if delay is None:
        delay = 5.0

    if verbose is None:
        verbose = False

    if verbose:
        logger = logging.getLogger(COMMON_LOGGER)
        logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        logger.addHandler(ch)

    def onepanel_push_wrapper():
        if not os.path.exists(path):
            logging.getLogger(COMMON_LOGGER).info('Waiting until path {} exists'.format(path))
            # Wait until path exists
            return

        logging.getLogger(COMMON_LOGGER).info('Doing command')
        jobs.upload_output_with_parameters(ctx, job_uid, project_uid, project_account_uid, source_directory=path, destination=destination)

    timer = Timer(delay, onepanel_push_wrapper)
    timer.start()
    timer.join()


from onepanel.models.api_json import APIJSON


class DatasetMountIdentifier(APIJSON):
    def __init__(self, account_uid=None, dataset_uid=None, version=None, destination=None):
        self.account_uid = account_uid
        self.dataset_uid = dataset_uid
        self.set_version(version)
        self.destination = destination

    def set_source(self, source):
        """Source is expected to be account_uid/dataset_uid"""
        source_parts = source.split('/')
        self.account_uid = source_parts[0]
        self.dataset_uid = source_parts[1]

    @classmethod
    def from_simple_json(cls, dct):
        identifier = cls()

        identifier.set_source(dct['source'])
        identifier.set_version(dct['version'])
        identifier.destination = dct['destination']

        return identifier

    def set_version(self, version):
        if version is None:
            self.version = None
            return
            
        self.version = int(version)

    def api_json(self):
        return {
            'accountUID': self.account_uid,
            'datasetUID': self.dataset_uid,
            'datasetVersion': self.version,
            'alias': self.destination
        }

    def __str__(self):
        return "{{accountUID:{}\ndatasetUID:{}\nversion:{}\nalias:{}}}".format(self.account_uid, self.dataset_uid, self.version, self.destination)




""" Command line interface for the OnePanel Machine Learning platform

'Login' commands group.
"""

import json
from functools import wraps

import click


@click.command(help="Login with email and password.")
@click.option('--email', prompt='Email')
@click.option('--password', prompt='Password', hide_input=True)
@click.pass_context
def login(ctx, email, password):
    conn = ctx.obj['connection']
    url = conn.URL + '/sessions'
    data = {
        'email': email,
        'password': password,
        'sessions': [
            {'device': 'cli'}
        ],
        'account': {}
    }

    r = conn.put(url, data=json.dumps(data))
    if r.status_code == 200:
        data = r.json()
        conn.save_credentials(data)
    elif r.status_code == 401 or r.status_code == 422:
        print('Incorrect username or password')
    else:
        print('Error: {}'.format(r.status_code))


@click.command(help="Pass in the gitlab token to login.",
               name="login-with-token",
               hidden=True
               )
@click.argument('token')
@click.pass_context
def login_with_token(ctx, token):
    conn = ctx.obj['connection']
    url = conn.URL + '/sessions'
    data = {
        'sessions': [
            {
                'token': token,
                'device': 'cli'
            }
        ]
    }

    r = conn.put(url, data=json.dumps(data))
    if r.status_code == 200:
        data = r.json()
        conn.save_credentials(data)
    elif r.status_code == 401 or r.status_code == 422:
        print('Invalid token')
    else:
        print('Error: {}'.format(r.status_code))


@click.command(help="Logs the current user out.",
               name="logout")
@click.pass_context
def logout(ctx):
    conn = ctx.obj['connection']
    conn.clear_credentials()

    click.echo('You are now logged out')


def login_required(func):
    """ Decorator that checks if the session is opened

    The decorator checks if the session is opened based on current context. Therefore please put the decorator
    after @click.pass_context decorator. For example:
        @click.command('hello')
        @click.option('--format')
        @click.pass_context
        @login_required
        def hello(ctx, format):
            ...
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if len(args) == 0:
            print('There is no context in the command. Please add @click.pass_context decorator')
            return

        ctx = args[0]
        conn = ctx.obj['connection']
        if conn.user_uid and conn.gitlab_impersonation_token:
            return func(*args, **kwargs)
        else:
            print('You are not logged in, '
                'log in by typing `onepanel login` at command line.')
    return wrapper



import click
import datetime

class DurationParamType(click.ParamType):
    name = 'duration'

    def convert(self, value, param, ctx):
        """Accepts dd:hh:mm:ss or a sub portion like mm:ss or just ss"""
        try:
            parts = value.split(':')
            parts.reverse()

            total_seconds = 0
            
            multiplier = 1
            for part in parts:
                total_seconds += int(part) * multiplier
                multiplier *= 60

            return datetime.timedelta(seconds=total_seconds)
        except ValueError:
            self.fail('%s is not a valid duration' % value, param, ctx)

DURATION = DurationParamType()